#############################################################################
# utils.tcl Copyright 2004, 2005 r0lZ
# Procedures related to the Macros and Info menus.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

namespace eval ::macros {
namespace export macros_menu info_menu
namespace import ::readdvd::*

#############################################################################
## macros_menu
## build the Macros menu
proc macros_menu {menuname} {
	menu $menuname
	$menuname add command -label "Kill PGC Playback" -command {::macros::kill_pgc_playback} -accelerator "Ctrl+K"
	bind . <Control-k> ::macros::kill_pgc_playback
	$menuname add command -label "Enable all operations" -command {::macros::enable_all_operations}
	$menuname add command -label "Reset Used GPRMs to 0" -command {::macros::reset_used_gprms}
	$menuname add separator
	$menuname add command -label "Jump to PGC upon DVD insert" -command {::macros::jump2pgc}
	$menuname add separator
	$menuname add command -label "Import First-Play (startup) Clip" -command {::macros::import_fp_clip}
	$menuname add command -label "Import Title's Intro Clip" -command {::macros::import_title_intro_clip}
	$menuname add command -label "Import Title's Closing Clip" -command {::macros::import_title_closing_clip}
	$menuname add separator
	$menuname add command -label "DVDShrink Streams Remapping" -command {::macros::check_streams_remapping}
	$menuname add command -label "DVDShrink Remote Buttons" -command {::macros::title_but_next_title}
	$menuname add separator
	$menuname add command -label "Fix number of streams in VMGM/VTSI_MAT tables" -command {::streams::verify_num_streams}
}

#############################################################################
## info_menu
## build the Info menu
proc info_menu {menuname} {
	menu $menuname
	$menuname add command -label "PGC" -command {::macros::show_current_pgc_as_text}
	$menuname add separator
	$menuname add command -label "Title Play Map table" -command ::srptinfo::srptinfo_dump
	$menuname add separator
	$menuname add command -label "DVD commands" -command ::main::dump_DVD
	$menuname add command -label "PGCs params and Cell Tables" -command ::pgced::dump_all_pgcs
	$menuname add command -label "Menu Buttons commands" -command {::menubuttons::show_menu_buttons_info}
	$menuname add command -label "Go to calling command" -command {::utils::pgc_crossrefs} -accelerator "Ctrl+G"
	bind . <Control-g> ::utils::pgc_crossrefs
	$menuname add command -label "Calls cross references" -command {::utils::show_crossrefs} -accelerator "Ctrl+H"
	bind . <Control-h> ::utils::show_crossrefs
	$menuname add command -label "Streams" -command ::streams::get_streams_info
	$menuname add separator
	$menuname add command -label "Find Unused GPRMs" -command {::macros::find_unused_gprms false}
}


#############################################################################
## Import VTS Title and automatically play it when the DVD play for the first time (PF-PGC)
proc import_fp_clip {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set utiltitle "Import first-play clip"

	# find a free GPRM
	set freegprm [lindex [find_unused_gprms true] end]
	if {$freegprm == ""} {
		set txt "This macro need an unused GPRM to work."
		set txt "$txt\nThere are no remaining free GPRMs!\nSorry."
		tk_messageBox -title $utiltitle -icon warning -parent . -message $txt
		return
	}
	set freegprm [format "%02X" $freegprm]

	### GUI
	if {![info exists ::macros::startupclip_stilltime]} {
		set ::macros::startupclip_stilltime -1		;# still time value: 0, # secs, or 255
		set ::macros::startupclip_stilltime_secs 10	;# used only in GUI
	}

	if {[winfo exists .importclip]} {destroy .importclip}
	set w .importclip
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w $utiltitle
	if {$::config(tk) != "aqua"} {wm transient $w .}
	wm protocol $w WM_DELETE_WINDOW {set ::macros::startupclip_done 0}
	bind $w <Escape> "set ::macros::startupclip_done 0"

	set txt "This macro will import a new titleset title,"
	set txt "$txt\nand modify the DVD to let the imported clip"
	set txt "$txt\nautomatically play when the DVD is starting."
	label $w.lab -text $txt
	pack  $w.lab -padx 2

	frame $w.f1 -bd 2 -relief groove
	pack  $w.f1 -padx 4 -pady 2 -fill x
	set txt "You will be prompted for the titleset to import when you select"
	set txt "$txt\nthe Continue button.  Select a VTS with only ONE title!"
	label $w.f1.lab -text $txt -anchor w -justify left
	pack  $w.f1.lab -side left

	frame $w.f2 -bd 2 -relief groove
	pack  $w.f2 -padx 4 -pady 2 -fill x
	set txt "If the imported clip is a still frame,"
	set txt "$txt\nplease select his Still Time (pause duration)"
	label $w.f2.lab -text $txt -anchor w -justify left
	pack  $w.f2.lab -side top -anchor w

	frame $w.f2.f
	pack  $w.f2.f -fill x
	radiobutton $w.f2.f.b2 -text "Display still frame" \
			-variable ::macros::startupclip_stilltime -value -1 -anchor w
	pack  $w.f2.f.b2 -side left
	entry $w.f2.f.e -textvariable ::macros::startupclip_stilltime_secs -width 4 \
			-validate all -validatecommand "::ced::value_entry_validate 0 254 {%d} {%P}"
	pack  $w.f2.f.e -side left
	label $w.f2.f.l -text "seconds (max 254)"
	pack  $w.f2.f.l -side left
	radiobutton $w.f2.b1 -text "Don't pause: it's an animated clip" \
			-variable ::macros::startupclip_stilltime -value 0 -anchor w
	pack  $w.f2.b1 -fill x

	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -anchor e
	button $w.fb.ok -text "Continue" -width 10 -command {set ::macros::startupclip_done 1}
	pack   $w.fb.ok -side left -padx 2
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::macros::startupclip_done 0}
	pack   $w.fb.cancel -side left -padx 2

	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
	wm maxsize $w [winfo reqwidth $w] [winfo reqheight $w]
	set wx [expr {[winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)}]
	set wy [expr {[winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25}]
	if {$wx < 10} {set wx 10}
	if {$wy < 22} {set wy 22}
	wm geometry $w +$wx+$wy
	wm deiconify $w
	focus $w
	grab set $w

	tkwait variable ::macros::startupclip_done
	grab release $w
	destroy $w
	if {!$::macros::startupclip_done} return

	set startupclip_stilltime $::macros::startupclip_stilltime
	if {$startupclip_stilltime == -1} {
		set startupclip_stilltime $::macros::startupclip_stilltime_secs
		if {$startupclip_stilltime == ""} {set startupclip_stilltime 0}
	}

	#################### here we go ##################
	show_msg "Importing Clip to be automatically played..."

	### import the new VTS with the intro clip
	set startupclip_title_num [::utils::import_VTST]
	if {$startupclip_title_num == 0} return

	show_msg "Modifying DVD commands..."

	### Modify the FP-PGC pre commands to jump to our newly created VMGM menu
	### get the original FP-PGC pre cmds
	set cmds $::pgcs(0,0)
	set idx2 [expr {[lsearch $cmds {* post}] - 1}]
	# we are adding 4 commands, so increment the GOTOs by 4 in the original commands
	for {set i 1} {$i <= $idx2} {incr i} {
		set cmd [lindex $cmds $i]
		set cmdchk "[string range $cmd 0 1][string range $cmd 4 4]"
		if {$cmdchk == "001" || $cmdchk == "003"} {
			set ln [expr {4 + [x2d [string range $cmd 21 22]]}]
			lset cmds $i "[string range $cmd 0 19] [format %02X $ln]"
		}
	}
	set ::pgcs(0,0) [linsert $cmds 1 \
			"00 A1 00 $freegprm 00 01 00 04" "71 00 00 $freegprm 00 01 00 00" \
			"30 02 00 00 00 [format "%02X" $startupclip_title_num] 00 00" \
			 {00 00 00 00 00 00 00 00}]
	 ;# If <free gprm> == 1 Goto 4; Set <free GPRM> = 1; Jump to title <new_title>; NOP

	### Modify the new Clip's post commands to jump back to the FP-PGC
	set ::pgcs($::pgcs(numvts),1) {{* pre} {* post} {30 08 00 00 01 00 00 00} {* cell}}

	### Allow to use the Next Chapter button to skip to the end of the imported clip
	set_int ::pgcs($::pgcs(numvts),1,header) 0x9C 1 2

	### create a new Root menu for the imported title that jump back to the FP-PGC
	set cmds {"* pre" "30 06 00 00 00 00 00 00" "* post" "* cell"} ;# JumpSS FP
	::utils::new_root_menu $::pgcs(numvts) true
	set ::menus($::pgcs(numvts),1,1) $cmds

	### Set the imported clip still time
	set_int ::pgcs($::pgcs(numvts),1,cellpbtable) \
			[expr {(($::pgcs($::pgcs(numvts),1,numcells) - 1) * 24) + 2}] \
			$startupclip_stilltime 1
	
	### Remove original cell commands numbers from the cell table
	for {set c 0} {$c < $::pgcs($::pgcs(numvts),1,numcells)} {incr c} {
		set_int ::pgcs($::pgcs(numvts),1,cellpbtable) [expr {($c * 24) + 3}] 0 1
	}

	### refresh interface
	::main::init_tree_menu
	::main::display_pgc pgc 0 0

	show_msg "Clip successfully appended as Title $startupclip_title_num."
	return
}

#############################################################################
## Import VTS Title and automatically play it before current title playback
proc import_title_intro_clip {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set utiltitle "Import title intro clip"

	### remember currently selected title
	set indexlst [split [lindex $::editor_cmds_from 1] ","]
	set origtitle_vts [lindex $indexlst 0]	;# VTS of original title
	set origtitle_ttn [lindex $indexlst 1]	;# TTN of original title

	if {[lindex $::editor_cmds_from 0] == "menu" || $origtitle_vts == 0} {
		set txt "You must select the VTST Title before"
		set txt "$txt\nwhich you want to play the clip"
		set txt "$txt\nbefore calling this macro."
		tk_messageBox -title $utiltitle -icon error -parent . -message $txt
		return
	}

	# find a free GPRM
	set freegprm [lindex [find_unused_gprms true] end]
	if {$freegprm == ""} {
		set txt "This macro need an unused GPRM to work."
		set txt "$txt\nThere are no remaining free GPRMs!\nSorry."
		tk_messageBox -title $utiltitle -icon warning -parent . -message $txt
		return
	}
	set freegprm [format "%02X" $freegprm]

	### GUI
	if {![info exists ::macros::startclip_stilltime]} {
		set ::macros::startclip_stilltime -1		;# still time value: 0, # secs, or 255
		set ::macros::startclip_stilltime_secs 10	;# used only in GUI
	}

	if {[winfo exists .importclip]} {destroy .importclip}
	set w .importclip
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w $utiltitle
	if {$::config(tk) != "aqua"} {wm transient $w .}
	wm protocol $w WM_DELETE_WINDOW {set ::macros::startclip_done 0}
	bind $w <Escape> "set ::macros::startclip_done 0"

	set txt "This macro will import a new titleset title,"
	set txt "$txt\nand modify the DVD to let the imported clip"
	set txt "$txt\nautomatically play before the currently selected title"
	set txt "$txt\n([$::widget(main_cmd_lab) cget -text])."
	label $w.lab -text $txt
	pack  $w.lab -padx 2

	frame $w.f1 -bd 2 -relief groove
	pack  $w.f1 -padx 4 -pady 2 -fill x
	set txt "You will be prompted for the titleset to import when you select"
	set txt "$txt\nthe Continue button.  Select a VTS with only ONE title!"
	label $w.f1.lab -text $txt -anchor w -justify left
	pack  $w.f1.lab -side left

	frame $w.f2 -bd 2 -relief groove
	pack  $w.f2 -padx 4 -pady 2 -fill x
	set txt "If the imported clip is a still frame,"
	set txt "$txt\nplease select his Still Time (pause duration)"
	label $w.f2.lab -text $txt -anchor w -justify left
	pack  $w.f2.lab -side top -anchor w

	frame $w.f2.f
	pack  $w.f2.f -fill x
	radiobutton $w.f2.f.b2 -text "Display still frame" \
			-variable ::macros::startclip_stilltime -value -1 -anchor w
	pack  $w.f2.f.b2 -side left
	entry $w.f2.f.e -textvariable ::macros::startclip_stilltime_secs -width 4 \
			-validate all -validatecommand "::ced::value_entry_validate 0 254 {%d} {%P}"
	pack  $w.f2.f.e -side left
	label $w.f2.f.l -text "seconds (max 254)"
	pack  $w.f2.f.l -side left
	radiobutton $w.f2.b1 -text "Don't pause: it's an animated clip" \
			-variable ::macros::startclip_stilltime -value 0 -anchor w
	pack  $w.f2.b1 -fill x

	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -anchor e
	button $w.fb.ok -text "Continue" -width 10 -command {set ::macros::startclip_done 1}
	pack   $w.fb.ok -side left -padx 2
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::macros::startclip_done 0}
	pack   $w.fb.cancel -side left -padx 2

	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
	wm maxsize $w [winfo reqwidth $w] [winfo reqheight $w]
	set wx [expr {[winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)}]
	set wy [expr {[winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25}]
	if {$wx < 10} {set wx 10}
	if {$wy < 22} {set wy 22}
	wm geometry $w +$wx+$wy
	wm deiconify $w
	focus $w
	grab set $w

	tkwait variable ::macros::startclip_done
	grab release $w
	destroy $w
	if {!$::macros::startclip_done} return

	set startclip_stilltime $::macros::startclip_stilltime
	if {$startclip_stilltime == -1} {
		set startclip_stilltime $::macros::startclip_stilltime_secs
		if {$startclip_stilltime == ""} {set startclip_stilltime 0}
	}

	#################### here we go ##################
	show_msg "Importing Clip to be automatically played..."

	# Title number of the original title
	set origtitle_titlenum $::pgcs($origtitle_vts,$origtitle_ttn,Title)

	### import the new VTS with the intro clip
	set startclip_title_num [::utils::import_VTST]
	if {$startclip_title_num == 0} return

	show_msg "Modifying DVD commands..."

	### create a new VMGM PGC that jumps to our new clip
	set newcmd "30 02 00 00 00 [format "%02X" $startclip_title_num] 00 00"	;# Jump_TT <new title>
	set cmds [list {* pre} $newcmd {* post} {* cell}]
	if {$::menus(0,numlus) == 0} {
		# we need to create a new menu for the VMGM
		::utils::new_title_menu true
		::utils::set_menu_type 0 1 1 0	;# change the new PGC menu type from Title to none
		set ::menus(0,1,1) $cmds
		set new_vmgm_pgc_jump_new 1
	} else {
		# create new PGC
		set new_vmgm_pgc_jump_new [clip_create_VMG_menu $cmds]
	}

	### create a new VMGM PGC that jumps to the original title
	set newcmd1 "71 00 00 $freegprm 00 01 00 00"	;# Set <free GPRM> = 1
	set newcmd2 "30 02 00 00 00 [format "%02X" $origtitle_titlenum] 00 00"	;# Jump_TT <orig title>
	set cmds [list {* pre} $newcmd1 $newcmd2 {* post} {* cell}]
	set new_vmgm_pgc_jump_orig [clip_create_VMG_menu $cmds]


	### Modify the original Title's pre commands to jump to our newly created VMGM menu
	### get the original title's pre cmds
	set cmds $::pgcs($origtitle_vts,$origtitle_ttn)
	set idx2 [expr {[lsearch $cmds {* post}] - 1}]
	# we are adding 4 commands, so increment the GOTOs by 4 in the original commands
	for {set i 1} {$i <= $idx2} {incr i} {
		set cmd [lindex $cmds $i]
		set cmdchk "[string range $cmd 0 1][string range $cmd 4 4]"
		if {$cmdchk == "001" || $cmdchk == "003"} {
			set ln [expr {4 + [x2d [string range $cmd 21 22]]}]
			lset cmds $i "[string range $cmd 0 19] [format %02X $ln]"
		}
	}
	set digits01 [format "%02X" [expr {$new_vmgm_pgc_jump_new / 0X100}]]
	set digits23 [format "%02X" [expr {$new_vmgm_pgc_jump_new & 0XFF}]]
	set ::pgcs($origtitle_vts,$origtitle_ttn) [linsert $cmds 1 \
			"00 A1 00 $freegprm 00 01 00 03" "30 08 $digits01 $digits23 01 C0 00 00" \
			"71 00 00 $freegprm 00 00 00 00" {00 00 00 00 00 00 00 00}]
	 ;# If <free gprm> == 1 Goto 3; CallSS-VMGM-pgc <new_1>; Set <free GPRM> = 0; NOP

	### Modify the new Clip's commands to jump to the new VMGM PGC 2
	set cmds {{* pre} {* post}}
	set digits01 [format "%02X" [expr {$new_vmgm_pgc_jump_orig / 0X100}]]
	set digits23 [format "%02X" [expr {$new_vmgm_pgc_jump_orig & 0XFF}]]
	lappend cmds "30 08 $digits01 $digits23 01 C0 00 00" ;# CallSS-VMGM-pgc <new_2>
	lappend cmds {* post}

	set ::pgcs($::pgcs(numvts),1) $cmds

	### Allow to use the Next Chapter button to skip to the end of the imported clip
	set_int ::pgcs($::pgcs(numvts),1,header) 0x9C 1 2

	# create a new Root menu for the imported title, if the original title has a root menu
	set original_root_menu_pgc 0
	if {$::menus($origtitle_vts,numlus) > 0} {
		for {set i 1} {$i <= $::menus($origtitle_vts,1,numpgcs)} {incr i} {
			if {$::menus($origtitle_vts,1,$i,mask) == 131} {
				# create a new VMGM PGC to jump to the original Root menu
				set cmd "30 06 00 [format "%02X" $origtitle_ttn] [format "%02X" $origtitle_vts] 83 00 00"
				;# JumpSS to Root menu of <new VTS> <new TTN>
				set cmds [list "* pre" $cmd "* post" "* cell"]
				set root_jump_pgc [clip_create_VMG_menu $cmds]

				# and create the imported VTS's root menu
				::utils::new_root_menu $::pgcs(numvts) true

				set digits01 [format "%02X" [expr {$::menus(0,1,numpgcs) / 0X100}]]
				set digits23 [format "%02X" [expr {$::menus(0,1,numpgcs) & 0XFF}]]
				set cmds [list "* pre" "30 06 $digits01 $digits23 00 C0 00 00" "* post" "* cell"]
				;# CallSS-VMGM-pgc <new root jump pgc>
				set ::menus($::pgcs(numvts),1,1) $cmds

				break
			}
		}
	}

	### Set the imported clip still time
	set_int ::pgcs($::pgcs(numvts),1,cellpbtable) \
			[expr {(($::pgcs($::pgcs(numvts),1,numcells) - 1) * 24) + 2}] \
			$startclip_stilltime 1

	### Remove original cell commands numbers from the cell table
	for {set c 0} {$c < $::pgcs($::pgcs(numvts),1,numcells)} {incr c} {
		set_int ::pgcs($::pgcs(numvts),1,cellpbtable) [expr {($c * 24) + 3}] 0 1
	}

	### refresh interface
	::main::init_tree_menu
	::main::display_pgc pgc $origtitle_vts $origtitle_ttn

	show_msg "Clip successfully appended as Title $startclip_title_num."
	return
}

#############################################################################
## Import VTS Title and automatically play it when current title playback is finished
proc import_title_closing_clip {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set utiltitle "Import title closing clip"

	### remember currently selected title
	set indexlst [split [lindex $::editor_cmds_from 1] ","]
	set origtitle_vts [lindex $indexlst 0]	;# VTS of original title
	set origtitle_ttn [lindex $indexlst 1]	;# TTN of original title

	if {[lindex $::editor_cmds_from 0] == "menu" || $origtitle_vts == 0} {
		set txt "You must select the VTST Title at the end"
		set txt "$txt\nof which you want to play the clip before"
		set txt "$txt\ncalling this macro."
		show_warning -title $utiltitle -icon error -parent . -message $txt
		return
	}

	### GUI
	if {![info exists ::macros::endclip_stilltime]} {
		set ::macros::endclip_stilltime -1		;# still time value: 0, # secs, or 255
		set ::macros::endclip_finish_mode 3		;# 0=exit, 1=FP-PGC, 2=Clone 3=VTS menu
		set ::macros::endclip_stilltime_secs 10	;# used only in GUI
	}

	if {[winfo exists .importclip]} {destroy .importclip}
	set w .importclip
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w $utiltitle
	if {$::config(tk) != "aqua"} {wm transient $w .}
	wm protocol $w WM_DELETE_WINDOW {set ::macros::endclip_done 0}
	bind $w <Escape> "set ::macros::endclip_done 0"

	set txt "This macro will import a new titleset title,"
	set txt "$txt\nand modify the DVD to let the imported clip"
	set txt "$txt\nautomatically play when the currently selected title"
	set txt "$txt\n([$::widget(main_cmd_lab) cget -text]) has finished playing."
	set txt "$txt\nThis may be used to show an \"Insert Disc 2\" still frame,"
	set txt "$txt\nor to add a featurette from a bonus disc after the main title."
	label $w.lab -text $txt
	pack  $w.lab -padx 2

	frame $w.f1 -bd 2 -relief groove
	pack  $w.f1 -padx 4 -pady 2 -fill x
	set txt "You will be prompted for the titleset to import when you select"
	set txt "$txt\nthe Continue button.  Select a VTS with only ONE title!"
	label $w.f1.lab -text $txt -anchor w -justify left
	pack  $w.f1.lab -side left

	frame $w.f2 -bd 2 -relief groove
	pack  $w.f2 -padx 4 -pady 2 -fill x
	set txt "If the imported clip is a still frame,"
	set txt "$txt\nplease select his Still Time (pause duration)"
	label $w.f2.lab -text $txt -anchor w -justify left
	pack  $w.f2.lab -side top -anchor w

	frame $w.f2.f
	pack  $w.f2.f -fill x
	radiobutton $w.f2.f.b2 -text "Display still frame" \
			-variable ::macros::endclip_stilltime -value -1 -anchor w \
			-command "::macros::validate_endclip_buttons $w"
	pack  $w.f2.f.b2 -side left
	entry $w.f2.f.e -textvariable ::macros::endclip_stilltime_secs -width 4 \
			-validate all -validatecommand "::ced::value_entry_validate 0 254 {%d} {%P}"
	pack  $w.f2.f.e -side left
	label $w.f2.f.l -text "seconds (max 254)"
	pack  $w.f2.f.l -side left
	radiobutton $w.f2.b3 -text "Display still frame forever" \
			-variable ::macros::endclip_stilltime -value 255 -anchor w \
			-command "::macros::validate_endclip_buttons $w"
	pack  $w.f2.b3 -fill x
	radiobutton $w.f2.b1 -text "Don't pause: it's an animated clip" \
			-variable ::macros::endclip_stilltime -value 0 -anchor w \
			-command "::macros::validate_endclip_buttons $w"
	pack  $w.f2.b1 -fill x

	frame $w.f3 -bd 2 -relief groove
	pack  $w.f3 -padx 4 -pady 2 -fill x
	set txt "When the end-clip has finished playing,"
	set txt "$txt\nwhat do you want to do?"
	label $w.f3.lab -text $txt -anchor w -justify left
	pack  $w.f3.lab -side top -anchor w
	radiobutton $w.f3.b0 -text "Return to the Root Menu of the original title." \
			-variable ::macros::endclip_finish_mode -value 3 -anchor w
	pack  $w.f3.b0 -fill x
	radiobutton $w.f3.b1 -text "Restart the DVD from the beginning. (FP-PGC)" \
			-variable ::macros::endclip_finish_mode -value 1 -anchor w
	pack  $w.f3.b1 -fill x
	radiobutton $w.f3.b2 -text "End the DVD playback. (Exit)" \
			-variable ::macros::endclip_finish_mode -value 0 -anchor w
	pack  $w.f3.b2 -fill x
	radiobutton $w.f3.b3 -text "Try to Clone the original title behavior." \
			-variable ::macros::endclip_finish_mode -value 2 -anchor w
	pack  $w.f3.b3 -fill x


	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -anchor e
	button $w.fb.ok -text "Continue" -width 10 -command {set ::macros::endclip_done 1}
	pack   $w.fb.ok -side left -padx 2
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::macros::endclip_done 0}
	pack   $w.fb.cancel -side left -padx 2

	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
	wm maxsize $w [winfo reqwidth $w] [winfo reqheight $w]
	set wx [expr {[winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)}]
	set wy [expr {[winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25}]
	if {$wx < 10} {set wx 10}
	if {$wy < 22} {set wy 22}
	wm geometry $w +$wx+$wy
	wm deiconify $w
	focus $w
	grab set $w

	### get the original title's post cmds
	set cmds $::pgcs($origtitle_vts,$origtitle_ttn)
	set idx1 [expr {[lsearch $cmds {* post}] + 1}]
	set idx2 [expr {[lsearch $cmds {* cell}] - 1}]
	if {$idx2 >= $idx1} {
		set origtitle_post_cmds [lrange $cmds $idx1 $idx2]
	} else {
		set origtitle_post_cmds ""
	}

	# find root menu PGC in original VTS
	set original_root_menu_pgc 0
	if {$::menus($origtitle_vts,numlus) > 0} {
		for {set i 1} {$i <= $::menus($origtitle_vts,1,numpgcs)} {incr i} {
			if {$::menus($origtitle_vts,1,$i,mask) == 131} {
				set original_root_menu_pgc $i
				break
			}
		}
	}

	# check if it is possible to 'clone the original title behavior'
	set cmds $origtitle_post_cmds
	set clone_ok true
	# convert VTST commands to VMGM commands
	if {$clone_ok} {
		set converted_origcmds [list "* pre"]
		foreach cmd $cmds {
			set cmd [string toupper $cmd]
			# check for SetSTN and SetNVTMR (illegal in the VMG)
			if {[lsearch -regexp $cmds {[45][12] .. .. .. .. .. .. ..}] >= 0} {
				set clone_ok false
				break
			}
			# check for links (can't be processed)
			if {[lsearch -regexp $cmds {20 .. .. .. .. .. .. ..}] >= 0} {
				set clone_ok false
				break
			}
			# CallSS-FP, -> JumpSS-FP and CallSS-VMGM-pgc -> JumpSS-VMGM-pgc
			if {[lsearch -regexp [list $cmd] {30 .8 .. .. .. [0C]. .. ..}] >= 0} {
				set cmd "[string range $cmd 0 3]6[string range $cmd 5 11]00[string range $cmd 14 end]"
			}
			# CallSS_VTSM -> JumpSS_VTSM
			if {[lsearch -regexp [list $cmd] {30 .8 .. .. .. 8. .. ..}] >= 0} {
				set cmd "[string range $cmd 0 3]6 00 [format "%02X" $origtitle_ttn] [format "%02X" $origtitle_vts] [string range $cmd 15 end]"
			}
			# CallSS-VMGM-menu, -> LinkPGCN (to menu pgc)
			if {[lsearch -regexp [list $cmd] {30 .8 .. .. .. 42 .. ..}] >= 0} {
				set pgc 0
				for {set i 1} {$i <= $::menus(0,1,numpgcs)} {incr i} {
					if {$::menus(0,1,$i,mask) == 130} {
						set pgc $i
						break
					}
				}
				if {$pgc == 0} {
					set clone_ok false
					break
				}
				set digit1  [format  "%1X" [expr {$pgc / 256}]]
				set digit23 [format "%02X" [expr {$pgc & 0XFF}]]
				set cmd "20 [string range $cmd 3 3]4 00 [string range $cmd 18 19] 00 [string range $cmd 21 22] 0$digit1 $digit23"
			}
			# JumpVTS_TT (ttn) -> JumpTT (title)
			if {[lsearch -regexp [list $cmd] {30 .3 .. .. .. .. .. ..}] >= 0} {
				set ttn [x2d [string range $cmd 15 16]]
				if {! [info exists ::pgcs($origtitle_vts,$ttn,Title)]} {
					set clone_ok false
					break
				} else {
					set title $::pgcs($origtitle_vts,$ttn,Title)
					set cmd "[string range $cmd 0 3]2[string range $cmd 5 14][format "%02X" $title][string range $cmd 17 end]"
				}
			}
			# JumpVTS_PTT (ptt) -> JumpTT (only if ptt==1)
			if {[lsearch -regexp [list $cmd] {30 .5 .. .. .. .. .. ..}] >= 0} {
				set ptt [x2d [string range $cmd 7 7][string range $cmd 9 10]]
				if {$ptt > 1} {
					set clone_ok false
					break
				}
				set ttn [x2d [string range $cmd 15 16]]
				if {! [info exists ::pgcs($origtitle_vts,$ttn,Title)]} {
					set clone_ok false
					break
				} else {
					set title $::pgcs($origtitle_vts,$ttn,Title)
					set cmd "[string range $cmd 0 3]2[string range $cmd 5 14][format "%02X" $title][string range $cmd 17 end]"
				}
			}
			lappend converted_origcmds $cmd
		}
		lappend converted_origcmds "* post" "* cell"
	}
	if {!$clone_ok} {
		set ::macros::clone_on_state "disabled"
		if {$::macros::endclip_finish_mode == 2} {
			set ::macros::endclip_finish_mode 3
		}
	} else {
		set ::macros::clone_on_state "normal"
	}
	if {$original_root_menu_pgc == 0} {
		set ::macros::root_on_state "disabled"
		if {$::macros::endclip_finish_mode == 3} {
			set ::macros::endclip_finish_mode 1
		}
	} else {
		set ::macros::root_on_state "normal"
	}

	validate_endclip_buttons $w
	if {$::macros::endclip_stilltime_secs == ""} {
		set ::macros::endclip_stilltime_secs 10
	}


	tkwait variable ::macros::endclip_done
	grab release $w
	destroy $w
	if {!$::macros::endclip_done} return

	set endclip_stilltime $::macros::endclip_stilltime
	if {$endclip_stilltime == -1} {
		set endclip_stilltime $::macros::endclip_stilltime_secs
		if {$endclip_stilltime == ""} {set endclip_stilltime 0}
	}
	set endclip_finish_mode $::macros::endclip_finish_mode
	if {$endclip_stilltime == 255} {
		set endclip_finish_mode 0
	}

	#################### here we go ##################
	# Variables initialized at this point:
	# origtitle_vts: VTS of original title
	# origtitle_ttn: TTN of original title
	# endclip_stilltime: still time value: 0, # secs, or 255
	# endclip_finish_mode: 0=exit, 1=FP-PGC, 2=Clone, 3=Root menu
	#
	# New variables to be initialized later:
	# origtitle_titlenum: Title number of the original title
	# endclip_title_num: Title number of new clip
	# new_vmgm_pgc_num: PGC number of the newly created VMGM PGC
	# origtitle_post_cmds: Original post commands of the original title

	show_msg "Importing Clip to be automatically played..."

	# Title number of the original title
	set origtitle_titlenum [subst \$::pgcs($origtitle_vts,$origtitle_ttn,Title)]

	### import the new VTS with the end clip
	set endclip_title_num [::utils::import_VTST]
	if {$endclip_title_num == 0} return

	### create a new VMGM PGC that jumps to our new clip
	show_msg "Modifying DVD commands..."
	set newcmd "30 02 00 00 00 [format "%02X" $endclip_title_num] 00 00"	;# Jump_TT
	set cmds [list {* pre} $newcmd {* post} {* cell}]
	if {$::menus(0,numlus) == 0} {
		# we need to create a new menu for the VMGM
		::utils::new_title_menu true
		::utils::set_menu_type 0 1 1 0	;# change the new PGC menu type from Title to none
		set ::menus(0,1,1) $cmds
		set new_vmgm_pgc_num 1
	} else {
		# create new PGC
		set new_vmgm_pgc_num [clip_create_VMG_menu $cmds]
	}

	### Modify the original Title's post commands to jump to our newly created VMGM menu
	set cmds $::pgcs($origtitle_vts,$origtitle_ttn)
	set idx1 [expr {[lsearch $cmds {* post}] + 1}]
	set idx2 [expr {[lsearch $cmds {* cell}] - 1}]
	# we are adding 2 commands, so increment the GOTOs by 2 in the original commands
	for {set i $idx1} {$i <= $idx2} {incr i} {
		set cmd [lindex $cmds $i]
		set cmdchk "[string range $cmd 0 1][string range $cmd 4 4]"
		if {$cmdchk == "001" || $cmdchk == "003"} {
			set ln [expr {2 + [x2d [string range $cmd 21 22]]}]
			lset cmds $i "[string range $cmd 0 19] [format %02X $ln]"
		}
	}
	set digits01 [format "%02X" [expr {$new_vmgm_pgc_num / 0X100}]]
	set digits23 [format "%02X" [expr {$new_vmgm_pgc_num & 0XFF}]]
	set ::pgcs($origtitle_vts,$origtitle_ttn) [linsert $cmds $idx1 \
			"30 08 $digits01 $digits23 01 C0 00 00" \
			{00 00 00 00 00 00 00 00}]	;# CallSS-VMGM-pgc ; NOP


	### Modify the new Clip's commands according to the finish mode
	set cmds {{* pre} {* post}}
	if {$endclip_finish_mode < 2} {
		# reset all used GPRMs to 0
		set unusedgprms [find_unused_gprms true]
		for {set r 0} {$r < 16} {incr r} {
			if {[lsearch $unusedgprms $r] == -1} {
				lappend cmds "71 00 00 [format "%02X" $r] 00 00 00 00"
			}
		}
	}
	# add finish command
	switch $endclip_finish_mode {
		0 {	# add 'exit'
			lappend cmds "30 01 00 00 00 00 00 00"
		}
		1 {	# add 'CallSS-FP, resume cell 1'
			lappend cmds "30 08 00 00 01 00 00 00"
		}
		2 -
		3 {	# clone original cmds OR Return to root menu
			# add CallSS-VMGM-pgc <new_pgc_num>, resume cell 1
			set new_vmgm_pgc_num2 [expr {$new_vmgm_pgc_num + 1}]
			set digits01 [format "%02X" [expr {$new_vmgm_pgc_num2 / 0X100}]]
			set digits23 [format "%02X" [expr {$new_vmgm_pgc_num2 & 0XFF}]]
			lappend cmds "30 08 $digits01 $digits23 01 C0 00 00"

			# if clone mode, create a new VMGM PGC for the original commands
			if {$endclip_finish_mode == 2} {
				clip_create_VMG_menu $converted_origcmds
			}
		}
	}
	lappend cmds {* cell}
	set ::pgcs($::pgcs(numvts),1) $cmds

	### Allow to use the Next Chapter button to skip to the end of the imported clip
	set_int ::pgcs($::pgcs(numvts),1,header) 0x9C 1 2

	# create a new Root menu for the imported title, if the original title has a root menu
	if {$original_root_menu_pgc > 0} {
		# create a new VMGM PGC to jump to the original Root menu
		set cmd "30 06 00 [format "%02X" $origtitle_ttn] [format "%02X" $origtitle_vts] 83 00 00"
		set cmds [list "* pre" $cmd "* post" "* cell"]
		set root_jump_pgc [clip_create_VMG_menu $cmds]

		# and create the imported VTS's root menu
		::utils::new_root_menu $::pgcs(numvts) true

		set digits01 [format "%02X" [expr {$::menus(0,1,numpgcs) / 0X100}]]
		set digits23 [format "%02X" [expr {$::menus(0,1,numpgcs) & 0XFF}]]
		set cmds [list "* pre" "30 06 $digits01 $digits23 00 C0 00 00" "* post" "* cell"]
		set ::menus($::pgcs(numvts),1,1) $cmds
	}

	### Set the imported clip still time
	set_int ::pgcs($::pgcs(numvts),1,cellpbtable) \
			[expr {(($::pgcs($::pgcs(numvts),1,numcells) - 1) * 24) + 2}] \
			$endclip_stilltime 1

	### Remove original cell commands numbers from the cell table
	for {set c 0} {$c < $::pgcs($::pgcs(numvts),1,numcells)} {incr c} {
		set_int ::pgcs($::pgcs(numvts),1,cellpbtable) [expr {($c * 24) + 3}] 0 1
	}

	### refresh interface
	::main::init_tree_menu
	::main::display_pgc pgc $origtitle_vts $origtitle_ttn

	show_msg "Clip successfully appended as Title $endclip_title_num."
	return
}
proc validate_endclip_buttons {w} {
	switch -- $::macros::endclip_stilltime {
		-1 {
			$w.f2.f.e configure -state normal
			$w.f3.b0  configure -state $::macros::root_on_state
			$w.f3.b1  configure -state normal
			$w.f3.b2  configure -state normal
			$w.f3.b3  configure -state $::macros::clone_on_state
			if {$::macros::endclip_stilltime_secs == ""} {
				set ::macros::endclip_stilltime_secs 10
			}
		}
		0 {
			$w.f2.f.e configure -state disabled
			$w.f3.b0  configure -state $::macros::root_on_state
			$w.f3.b1  configure -state normal
			$w.f3.b2  configure -state normal
			$w.f3.b3  configure -state $::macros::clone_on_state
		}
		255 {
			$w.f2.f.e configure -state disabled
			$w.f3.b0  configure -state disabled
			$w.f3.b1  configure -state disabled
			$w.f3.b2  configure -state disabled
			$w.f3.b3  configure -state disabled
		}
	}
}
proc clip_create_VMG_menu {cmds} {
	for {set lunum 1} {$lunum <= $::menus(0,numlus)} {incr lunum} {
		incr ::menus(0,$lunum,numpgcs)
		set pgcnum $::menus(0,$lunum,numpgcs)
		if {$lunum == 1} {
			set new_pgc_num $pgcnum
		}
		set ::menus(0,$lunum,$pgcnum,mask) 0
		set ::menus(0,$lunum,$pgcnum,masktext) ""
		set ::menus(0,$lunum,$pgcnum) $cmds
		set ::menus(0,$lunum,$pgcnum,header) [string repeat \000 236]
		if {[info exists ::pgcs(0,1,header)]} {
			set_int ::menus(0,$lunum,$pgcnum,header) 4 \
					[expr {[get_int ::pgcs(0,1,header) 4 4] & 0XC0}] 4
		} else {
			set_int ::menus(0,$lunum,$pgcnum,header) 4 64 4
		}
		set ::menus(0,$lunum,$pgcnum,numprograms) 0
		set ::menus(0,$lunum,$pgcnum,programmap) ""
		set ::menus(0,$lunum,$pgcnum,numcells) 0
		set ::menus(0,$lunum,$pgcnum,cellpbtable) ""
		set ::menus(0,$lunum,$pgcnum,cellpostable) ""
		set ::raw_sectors(0,VMGM_PGCI_UT) "$::raw_sectors(0,VMGM_PGCI_UT)[string repeat \000 8]"
	}
	return $new_pgc_num
}

#############################################################################
## Safety check for DVDShrink's Logical remapping of enabled streams option.
## Stored in: HKEY_CURRENT_USER\Software\DVD Shrink\DVD Shrink 3.2\Preferences\LogicalStreamRemapping
proc check_streams_remapping {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set title "Check DVDShrink Streams Remapping"

	set used false

	# Titles
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		for {set ttn 1} {$ttn <= $::pgcs($vts,numpgcs)} {incr ttn} {
			set used [search_setstn_gprm ::pgcs($vts,$ttn)]
			if {$used} break
		}
		if {$used} break
	}

	# Menus
	if {!$used} {
		for {set vts 1} {$vts <= $::menus(numvts)} {incr vts} {
			for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
				for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
					set used [search_setstn_gprm ::menus($vts,$lu,$pgc)]
					if {$used} break
				}
				if {$used} break
			}
			if {$used} break
		}
	}

	# Menu buttons
	if {!$used} {
		for {set vts 1} {$vts <= $::menus(numvts)} {incr vts} {
			set butcmds {}
			if {[info exists ::buttons::butsinfo(menu,$vts,errorcode)] && \
					$::buttons::butsinfo(menu,$vts,errorcode) == 0} {
				for {set vid 1} {$vid <= $::buttons::butsinfo(menu,$vts,numvobids)} {incr vid} {
					for {set cid 1} {$cid <= $::buttons::butsinfo(menu,$vts,$vid,numcellids)} {incr cid} {
						if {[info exists ::buttons::butsinfo(menu,$vts,$vid,$cid,numbuts)] && \
									$::buttons::butsinfo(menu,$vts,$vid,$cid,numbuts) > 0} {
							for {set group 1} {$group <= $::buttons::butsinfo(menu,$vts,$vid,$cid,numgroups)} {incr group} {
								for {set but 1} {$but <= $::buttons::butsinfo(menu,$vts,$vid,$cid,numbuts)} {incr but} {
									lappend butcmds $::buttons::butsinfo(menu,$vts,$vid,$cid,$group,$but,command)
								}
							}
						}
					}
				}
			}
			set used [search_setstn_gprm butcmds]
			if {$used} break
		}
		unset butcmds
	}

	# BOVs
	if {!$used} {
		for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
			set butcmds {}
			if {[info exists ::buttons::butsinfo(pgc,$vts,errorcode)] && \
					$::buttons::butsinfo(pgc,$vts,errorcode) == 0} {
				for {set vid 1} {$vid <= $::buttons::butsinfo(pgc,$vts,numvobids)} {incr vid} {
					for {set cid 1} {$cid <= $::buttons::butsinfo(pgc,$vts,$vid,numcellids)} {incr cid} {
						if {[info exists ::buttons::butsinfo(pgc,$vts,$vid,$cid,numbuts)] && \
									$::buttons::butsinfo(pgc,$vts,$vid,$cid,numbuts) > 0} {
							for {set group 1} {$group <= $::buttons::butsinfo(pgc,$vts,$vid,$cid,numgroups)} {incr group} {
								for {set but 1} {$but <= $::buttons::butsinfo(pgc,$vts,$vid,$cid,numbuts)} {incr but} {
									lappend butcmds $::buttons::butsinfo(pgc,$vts,$vid,$cid,$group,$but,command)
								}
							}
						}
					}
				}
			}
			set used [search_setstn_gprm butcmds]
			if {$used} break
		}
		unset butcmds
	}

	set txt "This macro checks if you may safely use the DVDShrink option\n"
	set txt "${txt}\"Logical remapping of enabled streams\" with the current DVD.\n\n"
	set txt "${txt}When this option is enabled, DVDShrink physically removes the\n"
	set txt "${txt}definitions of the stripped streams in the IFOs, and attempts to\n"
	set txt "${txt}change all SetSTN commands to point to the new string numbers.\n"
	set txt "${txt}However, it is unable to change the indirect SetSTN commands\n"
	set txt "${txt}(when a GPRM is used instead of a constant value), because the\n"
	set txt "${txt}number of the stream is not hardcoded in the command.\n"
	set txt "${txt}It is best to turn this option ON whenever possible, or there will be\n"
	set txt "${txt}references to non existing streams, which may confuse the player.\n"
	set txt "${txt}This means that it is better to use the logical remapping in Full Disc\n"
	set txt "${txt}mode but only if there are no indirect SetSTN command in the DVD.\n"
	set txt "${txt}Notes:\n"
	set txt "${txt}- This option is useless if you don't strip any stream.\n"
	set txt "${txt}- You may safely use it if you strip only the last streams (no gaps).\n"
	set txt "${txt}- The streams remapping option is ignored in Re-author mode.\n\n"
	set txt "${txt}Analysis results:\n\n"
	if {$used} {
		set txt "${txt}There are indirect SetSTN commands in this DVD.\n"
		set txt "${txt}(If you want to find them, search for \"\[41\" in the commands)\n"
		set txt "${txt}The logical remapping of enabled streams option should be OFF in\nFull Disc mode!"
		show_msg "DVDShrink's Logical remapping of enabled streams option is NOT safe."
		set rec "OFF"
		set default "no"
	} else {
		set txt "${txt}No indirect SetSTN commands found in this DVD.\n"
		set txt "${txt}The logical remapping of enabled streams option may safely be ON."
		show_msg "DVDShrink's Logical remapping of enabled streams option is safe."
		set rec "ON"
		set default "yes"
	}
	set txt "${txt}\n\nSummary for current DVD:\n\n"
	set txt "${txt}Recommended setting in Full Disc mode:  $rec  ($default)\n"

	set change false
	if {![catch {package require registry}]} {
		if {![catch {set cur [registry get \
				"HKEY_CURRENT_USER\\Software\\DVD Shrink\\DVD Shrink 3.2\\Preferences" \
				"LogicalStreamRemapping"]}]} {
			set change true
			if {$cur == 0} {
				set cur OFF
			} else {
				set cur ON
			}
		}
	}
	if {$change} {
		set txt "${txt}Current DVDShrink 3.2's remapping option:  $cur.\n\n"
		set txt "${txt}Allow Logical remapping of enabled streams in DVDShrink 3.2 ?"
		if {$default == "yes" && $cur == "ON"} {set default "cancel"}
		if {$default == "no" && $cur == "OFF"} {set default "cancel"}
		set r [tk_messageBox -title $title -icon info \
				-message $txt -parent . -type yesnocancel -default $default]
		if {$r == "yes"} {
			catch {registry set \
				"HKEY_CURRENT_USER\\Software\\DVD Shrink\\DVD Shrink 3.2\\Preferences" \
				"LogicalStreamRemapping" 1}
		} elseif {$r == "no"} {
			catch {registry set \
				"HKEY_CURRENT_USER\\Software\\DVD Shrink\\DVD Shrink 3.2\\Preferences" \
				"LogicalStreamRemapping" 0}
		}
	} else {
		tk_messageBox -title $title -icon info -message $txt -parent .
	}
	return $used
}
proc search_setstn_gprm {cmds_upvar} {
	upvar $cmds_upvar cmds
	if {[string first "\{41 " $cmds] == -1} {
		return false
	} else {
		return true
	}
}


#############################################################################
## add title button to skip to the next title on a DVD reauthored with DVDShrink
proc title_but_next_title {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set utiltitle "Navigation buttons for DVDShrink reauthored movies"
	if {!$::config(expert_mode)} {
		set txt "This macro work ONLY on DVDShrink reauthored DVDs!"
		set txt "$txt\n"
		set txt "$txt\nThe macro will modify the commands of (almost) every PGCs,"
		set txt "$txt\nand create the VMGM Title menu PGC, so that the user can"
		set txt "$txt\nuse the 'Title' button of the remote to skip to the next title."
		set txt "$txt\n"
		set txt "$txt\nThanks goes to 2COOL for the idea.  See this tread:"
		set txt "$txt\nhttp://forum.doom9.org/showthread.php?threadid=61238"
		set txt "$txt\n"
		set txt "$txt\nThe macro will also allow to use the 'Prev' and 'Next Chapter'"
		set txt "$txt\nbuttons to be used in respectively the first and last chapter"
		set txt "$txt\nof a title, and the 'Return' button to go back to the beginning"
		set txt "$txt\nof the current title, regardless of the current chapter."
		set txt "$txt\nThe last option allow to automatically loop back to the beginning"
		set txt "$txt\nof the DVD when the last title has finished playing."
		set txt "$txt\n"
		set txt "$txt\nWARNING: Only some basic checks are made to ensure that"
		set txt "$txt\nthe DVD is in the correct DVDShrink reauthored format."
		set txt "$txt\nUse this macro only with DVDShrink reauthored DVDs!"
		set txt "$txt\nALWAYS backup the DVD IFOs before using this macro!"
		set txt "$txt\n"
		set txt "$txt\nContinue ?"
		set r [tk_messageBox -title $utiltitle -icon question -message $txt \
				-parent . -type yesno -default no]
		if {$r == "no"} return
	}

	### init GUI defaults to reflect current state
	if {$::menus(0,numlus)==1 && $::menus(0,1,numpgcs)==1} {
		set ::macros::dvdshrinkbuttons_title 1
	} else {
		set ::macros::dvdshrinkbuttons_title 0
	}

	set ::macros::dvdshrinkbuttons_prev [get_int ::pgcs(1,1,header) 0x9E 2]
	if {$::macros::dvdshrinkbuttons_prev > 1} {set ::macros::dvdshrinkbuttons_prev 1}

	set ::macros::dvdshrinkbuttons_next [get_int ::pgcs(1,1,header) 0x9C 2]
	if {$::macros::dvdshrinkbuttons_next > 1} {set ::macros::dvdshrinkbuttons_next 1}

	set ::macros::dvdshrinkbuttons_return [get_int ::pgcs(1,1,header) 0xA0 2]
	if {$::macros::dvdshrinkbuttons_return > 1} {set ::macros::dvdshrinkbuttons_return 1}

	if {[lsearch [get_type_cmds post ::pgcs($::pgcs(numvts),1)] \
			{30 08 00 00 01 00 00 00}] >= 0} {	;# CallSS FP
		set ::macros::dvdshrinkbuttons_loop 1
	} else {
		set ::macros::dvdshrinkbuttons_loop 0
	}

	### GUI
	if {[winfo exists .dvdshrinkbuttons]} {destroy .dvdshrinkbuttons}
	set w .dvdshrinkbuttons
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w $utiltitle
	if {$::config(tk) != "aqua"} {wm transient $w .}
	wm protocol $w WM_DELETE_WINDOW {set ::macros::dvdshrinkbuttons_done 0}
	bind $w <Escape> "set ::macros::dvdshrinkbuttons_done 0"

	set txt "This macro allow to use the remote's buttons to navigate in DVDs reauthored with DVDShrink."
	label $w.lab -text $txt
	pack  $w.lab -padx 2 -pady 4

	frame $w.f1 -bd 2 -relief groove
	pack  $w.f1 -padx 4 -pady 2 -fill x
	checkbutton $w.f1.b2 -text "When in chapter 1, allow the 'Prev Chapter' button to go back to the beginning of the current title?" \
			-variable ::macros::dvdshrinkbuttons_prev
	pack  $w.f1.b2 -anchor w
	checkbutton $w.f1.b3 -text "When in last chapter, allow the 'Next Chapter' button to skip to the next title?" \
			-variable ::macros::dvdshrinkbuttons_next
	pack  $w.f1.b3 -anchor w
	checkbutton $w.f1.b6 -text "Always allow the 'Return' (GoUp) button to go back to the beginning of the current title?" \
			-variable ::macros::dvdshrinkbuttons_return
	pack  $w.f1.b6 -anchor w
	checkbutton $w.f1.b1 -text "Always allow the 'Title Menu' button to skip to the next title?" \
			-variable ::macros::dvdshrinkbuttons_title
	pack  $w.f1.b1 -anchor w
	checkbutton $w.f1.b4 -text "At the end of the last title, automatically loop back to the first title?" \
			-variable ::macros::dvdshrinkbuttons_loop
	pack  $w.f1.b4 -anchor w
	button $w.f1.b5 -text "Check all" -width 10 -command {
		set ::macros::dvdshrinkbuttons_title  1
		set ::macros::dvdshrinkbuttons_return 1
		set ::macros::dvdshrinkbuttons_prev   1
		set ::macros::dvdshrinkbuttons_next   1
		set ::macros::dvdshrinkbuttons_loop   1
	}
	pack  $w.f1.b5 -anchor w -padx 4 -pady 4


	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -anchor e
	button $w.fb.ok -text "Continue" -width 10 -command {set ::macros::dvdshrinkbuttons_done 1}
	pack   $w.fb.ok -side left -padx 2
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::macros::dvdshrinkbuttons_done 0}
	pack   $w.fb.cancel -side left -padx 2

	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
	wm maxsize $w [winfo reqwidth $w] [winfo reqheight $w]
	set wx [expr {[winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)}]
	set wy [expr {[winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25}]
	if {$wx < 10} {set wx 10}
	if {$wy < 22} {set wy 22}
	wm geometry $w +$wx+$wy
	wm deiconify $w
	focus $w
	grab set $w

	tkwait variable ::macros::dvdshrinkbuttons_done
	grab release $w
	destroy $w
	if {!$::macros::dvdshrinkbuttons_done} return

	if {$::macros::dvdshrinkbuttons_title} {
		### create or replace the VMGM menu
		set r "yes"
		if {$::menus(0,numlus)>0} {
			set txt "WARNING: There is already a VMGM Menu!"
			set txt "$txt\nIf you continue, it will be REPLACED by a new one!"
			set txt "$txt\n"
			set txt "$txt\nContinue anyway ?"
			set r [tk_messageBox -title $utiltitle -icon warning -message $txt \
					-parent . -type yesno -default yes]
		}
		if {$r == "yes"} {
			set ::raw_sectors(0,VMGM_PGCI_UT) \
					[readFile $::config(apppath)/lib/data/vmgm_pgci_ut_wo_vob.sec binary]

			parse_PGCI_UT ::raw_sectors(0,VMGM_PGCI_UT) 0 ::menus
			::main::init_tree_menu

			### for each VTS except last one, replace the pre command "Set gprm(0) = 0"
			### by "Set gprm(0) = <current VTS number>"
			for {set vts 1} {$vts < $::pgcs(numvts)} {incr vts} {
				set cmds [get_type_cmds pre ::pgcs($vts,1)]
				set idx [lsearch -regexp $cmds {71 00 00 00 .. .. 00 00}]
				if {$idx == -1} {
					set cmds [linsert $cmds 0 "71 00 00 00 00 [format %02X $vts] 00 00"]
				} else {
					lset cmds $idx "71 00 00 00 00 [format %02X $vts] 00 00"
				}
				set_type_cmds pre ::pgcs($vts,1) $cmds
			}
		}
	} else {
		### delete the VMGM menu
		set r "yes"
		if {$::menus(0,numlus)>0} {
			set txt "WARNING: The entire VMGM Menu will be DELETED!"
			set txt "$txt\n"
			set txt "$txt\nContinue anyway ?"
			set r [tk_messageBox -title $utiltitle -icon warning -message $txt \
					-parent . -type yesno -default yes]
			if {$r == "no"} return
		}
		if {$r == "yes"} {
			set ::raw_sectors(0,VMGM_PGCI_UT) ""
			set ::menus(0,numlus) 0
			::main::init_tree_menu

			### for each VTS except last one, replace the pre command
			### "Set gprm(0) = <current VTS number>" by "Set gprm(0) = 0"
			for {set vts 1} {$vts < $::pgcs(numvts)} {incr vts} {
				set cmds [get_type_cmds pre ::pgcs($vts,1)]
				set idx [lsearch -regexp $cmds {71 00 00 00 .. .. 00 00}]
				if {$idx != -1} {
					lset cmds $idx "71 00 00 00 00 00 00 00"
				}
				set_type_cmds pre ::pgcs($vts,1) $cmds
			}
		}
	}

	if {$::macros::dvdshrinkbuttons_loop} {
		### Last VTS: replace Exit by CallSS FP
		set vts $::pgcs(numvts)
		set cmds [get_type_cmds post ::pgcs($vts,1)]
		if {$cmds == ""} {
			set cmds {{30 08 00 00 01 00 00 00}}
		} else {
			set idx [lsearch $cmds {30 01 00 00 00 00 00 00}]	;# Exit
			if {$idx >= 0} {
				lset cmds $idx "30 08 00 00 01 00 00 00"	;# CallSS FP
			}
		}
		set_type_cmds post ::pgcs($vts,1) $cmds
	} else {
		### Last VTS: replace CallSS-FP by Exit
		set vts $::pgcs(numvts)
		set cmds [get_type_cmds post ::pgcs($vts,1)]
		if {$cmds == ""} {
			set cmds {{30 01 00 00 00 00 00 00}}
		} else {
			set idx [lsearch $cmds {30 08 00 00 01 00 00 00}]	;# CallSS FP
			if {$idx >= 0} {
				lset cmds $idx "30 01 00 00 00 00 00 00"	;# Exit
			}
		}
		set_type_cmds post ::pgcs($vts,1) $cmds
	}

	### Allow to use the Prev button, when pressed during Chapter 1) to go back to the
	### beginning of the title, and the Next Button to go directly to the end of the Title
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		if {$::macros::dvdshrinkbuttons_prev} {
			set_int ::pgcs($vts,1,header) 0x9E 1 2
		} else {
			set_int ::pgcs($vts,1,header) 0x9E 0 2
		}
		if {$::macros::dvdshrinkbuttons_next} {
			set_int ::pgcs($vts,1,header) 0x9C 1 2
		} else {
			set_int ::pgcs($vts,1,header) 0x9C 0 2
		}
		if {$::macros::dvdshrinkbuttons_return} {
			set_int ::pgcs($vts,1,header) 0xA0 1 2
		} else {
			set_int ::pgcs($vts,1,header) 0xA0 0 2
		}
	}

	### util done.  Refresh listbox.
	refresh_current_pgc
	::main::dvd_modified true false

	# if there is an old VMGM VOB, delete it or move it to backup
	::utils::delete_vob_with_confirm "VIDEO_TS.VOB" true
}

#############################################################################
## enable navigation buttons on the remote
proc nav_buttons {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}

	set curtype [lindex $::editor_cmds_from 0]
	set curvts  [lindex [split [lindex $::editor_cmds_from 1] ","] 0]

	if {![info exists ::macros::navbuttons_prev]} {
		set ::macros::navbuttons_prev   1
		set ::macros::navbuttons_next   1
		set ::macros::navbuttons_return 1
		set ::macros::navbuttons_menu	1
		set ::macros::navbuttons_menu_opt 1
		set ::macros::navbuttons_domain 1
	}

	### GUI
	set w .navbuttons
	if {[winfo exists $w]} {destroy $w}
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w "Enable Remote Navigation Buttons"
	if {$::config(tk) != "aqua"} {wm transient $w .}
	wm protocol $w WM_DELETE_WINDOW {set ::macros::navbuttons_done 0}
	bind $w <Escape> "set ::macros::navbuttons_done 0"

	frame $w.f1 -bd 2 -relief groove
	pack  $w.f1 -padx 4 -pady 4 -fill x
	checkbutton $w.f1.b2 -text "When in chapter 1, allow the 'Prev Chapter' button\nto go back to the beginning of the current PGC?" \
			-variable ::macros::navbuttons_prev -justify left
	pack  $w.f1.b2 -anchor w

	checkbutton $w.f1.b3 -text "When in last chapter, allow the 'Next Chapter' button\nto skip to the end of the PGC?" \
			-variable ::macros::navbuttons_next -justify left
	pack  $w.f1.b3 -anchor w

	checkbutton $w.f1.b6 -text "In the Title domain, allow the 'Return' (GoUp) button\nto go back to the beginning of the current PGC?" \
			-variable ::macros::navbuttons_return -justify left
	pack  $w.f1.b6 -anchor w

	checkbutton $w.f1.b4 -text "In the Menu domain, allow the 'Return' (GoUp) button" \
			-variable ::macros::navbuttons_menu -justify left \
			-command "if \$::macros::navbuttons_menu {set state normal} else {set state disabled}; $w.f1.f4.f.r1 configure -state \$state;  $w.f1.f4.f.r2 configure -state \$state"
	pack  $w.f1.b4 -anchor w
	frame $w.f1.f4
	pack  $w.f1.f4 -fill x
	frame $w.f1.f4.pad -width 20
	pack  $w.f1.f4.pad -side left
	frame $w.f1.f4.f
	pack  $w.f1.f4.f -fill x
	radiobutton $w.f1.f4.f.r1 -text "...to go back to the beginning of the current PGC?" \
			-variable ::macros::navbuttons_menu_opt -value 1 -justify left -bd 1
	pack  $w.f1.f4.f.r1 -anchor w
	radiobutton $w.f1.f4.f.r2 -text "...to resume to the last visited Title?" \
			-variable ::macros::navbuttons_menu_opt -value 0 -justify left -bd 1
	pack  $w.f1.f4.f.r2 -anchor w
	if {$::macros::navbuttons_menu} {set state normal} else {set state disabled}
	$w.f1.f4.f.r1 configure -state $state
	$w.f1.f4.f.r2 configure -state $state

	frame $w.f2 -bd 2 -relief groove
	pack  $w.f2 -padx 4 -pady 2 -fill x
	radiobutton $w.f2.r1 -variable ::macros::navbuttons_domain -value 1 \
			-text "Apply to current PGC only"
	pack  $w.f2.r1 -anchor w
	radiobutton $w.f2.r2 -variable ::macros::navbuttons_domain -value 2 \
			-text "Apply to all PGCs of the current domain"
	pack  $w.f2.r2 -anchor w
	radiobutton $w.f2.r3 -variable ::macros::navbuttons_domain -value 3 \
			-text "Apply to all Titles of the DVD"
	pack  $w.f2.r3 -anchor w
	radiobutton $w.f2.r4 -variable ::macros::navbuttons_domain -value 4 \
			-text "Apply to all PGCs of the VMGM menu"
	pack  $w.f2.r4 -anchor w

	if {$curtype == "pgc" && $curvts == 0} {$w.f2.r2 configure -state disabled}

	set txt "Notes: The buttons behavior are changed only when not"
	set txt "$txt\nalready used for another purpose, and are not changed"
	set txt "$txt\nfor 'dummy' menu PGCs (without video)."
	set txt "$txt\nThe 'Next Chapter' button is enabled only if there are post"
	set txt "$txt\ncommands in the PGC."
	set txt "$txt\nEnabling the buttons on menus may lead to strange results!"
	label $w.lab -text $txt -justify left
	pack  $w.lab -fill x -anchor w -padx 4

	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -anchor e
	button $w.fb.app -text "Apply" -width 10 -command {set ::macros::navbuttons_done 2}
	pack   $w.fb.app -side left -padx 2
	button $w.fb.ok  -text "OK" -width 10 -command {set ::macros::navbuttons_done 1}
	pack   $w.fb.ok  -side left -padx 2
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::macros::navbuttons_done 0}
	pack   $w.fb.cancel -side left -padx 2

	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
	wm maxsize $w [winfo reqwidth $w] [winfo reqheight $w]
	set wx [expr {[winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)}]
	set wy [expr {[winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25}]
	if {$wx < 10} {set wx 10}
	if {$wy < 22} {set wy 22}
	wm geometry $w +$wx+$wy
	wm deiconify $w
	focus $w

	set done 0
	while {! $done} {
		set ::macros::navbuttons_done -1
		tkwait variable ::macros::navbuttons_done
		if {$::macros::navbuttons_done==0} {
			destroy $w
			return
		}
		if {$::macros::navbuttons_done==1} {
			destroy $w
			set done 1
		}

		if {($curtype == "menu" && $::macros::navbuttons_domain!=3) \
				|| $::macros::navbuttons_domain==4} {
			set curlu  [lindex [split [lindex $::editor_cmds_from 1] ","] 1]
			set curpgc [lindex [split [lindex $::editor_cmds_from 1] ","] 2]
			switch $::macros::navbuttons_domain {
				1 {
					set startvts $curvts
					set endvts   $curvts
					set startlu  $curlu
					set endlu    $curlu
					set startpgc $curpgc
					set endpgc   $curpgc
				}
				2 {
					set startvts $curvts
					set endvts   $curvts
					set startlu  1
					set endlu    $::menus($curvts,numlus)
					set startpgc 1
					set endpgc   -1
				}
				4 {
					set startvts 0
					set endvts   0
					set startlu  1
					set endlu    $::menus(0,numlus)
					set startpgc 1
					set endpgc   -1
				}
			}
			for {set vts $startvts} {$vts <= $endvts} {incr vts} {
				for {set lu $startlu} {$lu <= $endlu} {incr lu} {
					if {$endpgc == -1} {
						set endpgc2 $::menus($vts,$lu,numpgcs)
					} else {
						set endpgc2 $endpgc
					}
					for {set pgc $startpgc} {$pgc <= $endpgc2} {incr pgc} {
						if {$::menus($vts,$lu,$pgc,numprograms)>0} {
							if {$::macros::navbuttons_prev   && \
									[get_int ::menus($vts,$lu,$pgc,header) 0x9E 2]==0} {
								set_int ::menus($vts,$lu,$pgc,header) 0x9E $pgc 2
							}
							if {$::macros::navbuttons_next   && \
									[get_int ::menus($vts,$lu,$pgc,header) 0x9C 2]==0 &&
									[num_type_cmds post ::menus($vts,$lu,$pgc)]>0} {
								set_int ::menus($vts,$lu,$pgc,header) 0x9C $pgc 2
							}
							if {$::macros::navbuttons_menu && \
									[get_int ::menus($vts,$lu,$pgc,header) 0xA0 2]==0} {
								if {$::macros::navbuttons_menu_opt} {
									set_int ::menus($vts,$lu,$pgc,header) 0xA0 $pgc 2
								} else {
									set_int ::menus($vts,$lu,$pgc,header) 0xA0 0xFFFF 2
								}
							}
						}
					}
				}
			}
		} else {
			set curpgc [lindex [split [lindex $::editor_cmds_from 1] ","] 1]
			switch $::macros::navbuttons_domain {
				1 {
					set startvts $curvts
					set endvts   $curvts
					set startpgc $curpgc
					set endpgc   $curpgc
				}
				2 {
					if {$curvts == 0} {
						bell
						return
					}
					set startvts $curvts
					set endvts   $curvts
					set startpgc 1
					set endpgc   $::pgcs($curvts,numpgcs)
				}
				3 {
					set startvts 1
					set endvts   $::pgcs(numvts)
					set startpgc 1
					set endpgc   -1
				}
			}
			for {set vts $startvts} {$vts <= $endvts} {incr vts} {
				if {$endpgc == -1} {
					set endpgc2 $::pgcs($vts,numpgcs)
				} else {
					set endpgc2 $endpgc
				}
				for {set pgc $startpgc} {$pgc <= $endpgc2} {incr pgc} {
					if {$::macros::navbuttons_prev   && \
							[get_int ::pgcs($vts,$pgc,header) 0x9E 2]==0} {
						set_int ::pgcs($vts,$pgc,header) 0x9E $pgc 2
					}
					if {$::macros::navbuttons_next   && \
							[get_int ::pgcs($vts,$pgc,header) 0x9C 2]==0 &&
							[num_type_cmds post ::pgcs($vts,$pgc)]>0} {
						set_int ::pgcs($vts,$pgc,header) 0x9C $pgc 2
					}
					if {$::macros::navbuttons_return && \
							[get_int ::pgcs($vts,$pgc,header) 0xA0 2]==0} {
						set_int ::pgcs($vts,$pgc,header) 0xA0 $pgc 2
					}
				}
			}
		}
		### util done.  Refresh listbox.
		refresh_current_pgc
		::main::dvd_modified true false
	}
	return
}


#############################################################################
## Kill PGC Playback
## Modify the PGC pre commands so that the PGC video is not played anymore.
proc kill_pgc_playback {{pgcvar_arg ""} {checkmode ask}} {
	if {! [info exists ::dvddir]} {
		bell
		return 0
	}
	if {$pgcvar_arg != ""} {
		set pgcvar $pgcvar_arg
	} else {
		set pgcvar ::[lindex $::editor_cmds_from 0]s([lindex $::editor_cmds_from 1])
		if {!$::config(expert_mode)} {
			set txt "This macro will inhibit the video playback"
			set txt "$txt\nof the current PGC."
			set txt "$txt\n\nA check is done to ensure that it is safe"
			set txt "$txt\nto kill playback on this particular PGC."
			set txt "$txt\n\nNote: The VOB file will not be modified."
			set txt "$txt\n"
			set txt "$txt\nContinue ?"
			set r [tk_messageBox -title "Kill PGC Playback" -icon warning -message $txt \
					-parent . -type yesno -default yes]
			if {$r == "no"} {
				return 0
			}
		}
	}
	set index [string trim [string range $pgcvar [string first "(" $pgcvar] end] "()"]
	set indexes [split $index ","]
	set numbuttons 0
	if {[string first "menu" $pgcvar] != -1} {
		set type "menu"
		set vts [lindex $indexes 0]
		set lu  [lindex $indexes 1]
		set pgc [lindex $indexes 2]
		set numcells $::menus($vts,$lu,$pgc,numcells)
		set pgclabel [::main::pgc_label "" menu $vts $lu $pgc]
		if {[info exists ::menus($index,numbuttons)]} {
			set numbuttons $::menus($index,numbuttons)
		}
	} else {
		set type "pgc"
		set vts [lindex $indexes 0]
		set lu  -1
		set pgc [lindex $indexes 1]
		set numcells $::pgcs($vts,$pgc,numcells)
		set pgclabel [::main::pgc_label "" pgc $vts $pgc]
		if {[info exists ::pgcs($index,numbuttons)]} {
			set numbuttons $::pgcs($index,numbuttons)
		}
	}
	if {$checkmode != "always"} {
		if {$numbuttons > 0} {
			if {$checkmode == "ask"} {
				set txt "As there are menu buttons in this PGC,"
				set txt "$txt\nit is not safe to kill playback in this case!"
				set txt "$txt\n"
				set txt "$txt\nContinue anyway?"
				set r [tk_messageBox -title "Kill PGC Playback" -icon warning -message $txt \
						-parent . -type yesno -default no]
				if {$r == "no"} {
					return 0
				}
			} elseif {$checkmode == "safely"} {
				return 0
			}
		}
		set safe [check_safe_kill_pb $type $vts $lu $pgc]
		if {$checkmode == "safely" && $safe != 1} {
			return 0
		}
		if {$checkmode == "ask" && $safe == 0} {
			set txt "As at least one cell-command jumps outside this PGC,"
			set txt "$txt\nit is not safe to kill playback in this case!"
			set txt "$txt\n"
			set txt "$txt\nContinue anyway?"
			set r [tk_messageBox -title "Kill PGC Playback" -icon warning -message $txt \
					-parent . -type yesno -default no]
			if {$r == "no"} {
				return 0
			}
		}
		if {$checkmode == "ask" && $safe == -1} {
			set txt "This PGC is a dummy PGC."
			set txt "$txt\nIt is useless to kill playback in this case."
			set txt "$txt\n"
			set txt "$txt\nContinue anyway?"
			set r [tk_messageBox -title "Kill PGC Playback" -icon question -message $txt \
					-parent . -type yesno -default no]
			if {$r == "no"} {
				return 0
			}
		}
		
	}
	# get pre-commands, and add a NOP command at the end, if needed and possible.
	set cmds [get_type_cmds pre $pgcvar]
	if {[lindex $cmds end]!={00 00 00 00 00 00 00 00}} {
		if {[expr {[num_type_cmds pre $pgcvar]+[num_type_cmds post $pgcvar]+[num_type_cmds cell $pgcvar]}] < 128 ||
				[num_type_cmds post $pgcvar]==0} {
			lappend cmds {00 00 00 00 00 00 00 00}	;# NOP
		}
	}
	set gotocmd [llength $cmds]	;# we will change the Breaks to Gotos to this line number
	set oktoblank true

	# search for Break commands
	set idxs [lsearch -regexp -all $cmds {00 .2 .. .. .. .. .. ..}]	;# Break
	foreach idx $idxs {
		set bcmd [lindex $cmds $idx]
		set gcmd [string range $bcmd 0 3]1[string range $bcmd 5 20][format %02X $gotocmd]
		lset cmds $idx $gcmd
	}
	# search for Link*Cell and Link*PG commands
	# replace them with gotos only if the Highlight parameter is NOT present.
	set idxs [lsearch -regexp -all $cmds {20 .1 .. .. .. .. .. 0[1-7]}]	;# Link*Cell and Link*PG
	foreach idx $idxs {
		set bcmd [lindex $cmds $idx]
		if {[string range $bcmd 18 19] == "00"} {
			set gcmd "00 [string range $bcmd 3 3]1 00 [string range $bcmd 9 16] 00 [format %02X $gotocmd]"
			lset cmds $idx $gcmd
		} else {
			set oktoblank false
		}
	}
	# search for LinkPTT, LinkPGN and LinkCN commands
	# replace them with gotos only if the Highlight parameter is NOT present.
	set idxs [lsearch -regexp -all $cmds {20 .[5-7] .. .. .. .. .. ..}]	;# LinkPTT, LinkPGN and LinkCN
	foreach idx $idxs {
		set bcmd [lindex $cmds $idx]
		if {[string range $bcmd 18 19] == "00"} {
			set gcmd "00 [string range $bcmd 3 3]1 00 [string range $bcmd 9 16] 00 [format %02X $gotocmd]"
			lset cmds $idx $gcmd
		} else {
			set oktoblank false
		}
	}

	# add post-commands after the NOP
	set postcmds [get_type_cmds post $pgcvar]
	# increment GOTOs
	set idxs [lsearch -regexp -all $postcmds {00 .[13] .. .. .. .. .. ..}]	;# GOTOs
	foreach idx $idxs {
		set cmd [lindex $postcmds $idx]
		set ln [expr {[x2d [lindex $cmd end]]+$gotocmd}]
		lset cmd end [format "%02X" $ln]
		lset postcmds $idx $cmd
	}
	# append post-cmds to pre-cmds
	set cmds [concat $cmds $postcmds]

	# replace pre-commands
	set_type_cmds pre $pgcvar $cmds

	# remove post commands if the total number of commands > 128
	if {[expr {[num_type_cmds pre $pgcvar]+[num_type_cmds post $pgcvar]+[num_type_cmds cell $pgcvar]}] > 128} {
		set_type_cmds post $pgcvar {}
		show_msg "Kill Playback WARNING: post commands removed!  (Max 128 commands in a PGC)"
		set oktoblank true
		bell
	}

	# reset PGC Still Time to 0
	if {$type == "menu"} {
		set_int ::menus($index,header) 0xA3 0 1
		for {set cell 1} {$cell <= $::menus($index,numcells)} {incr cell} {
			set cellidx [expr {($cell-1)*24}]
			# still time = 0
			set_int ::menus($index,cellpbtable) [expr {$cellidx+2}] 0 1
		}
	} else {
		set_int ::pgcs($index,header) 0xA3 0 1
		for {set cell 1} {$cell <= $::pgcs($index,numcells)} {incr cell} {
			set cellidx [expr {($cell-1)*24}]
			# still time = 0
			set_int ::pgcs($index,cellpbtable) [expr {$cellidx+2}] 0 1
		}
	}

	# remove Cells and Programs if type is menu.
	if {! $oktoblank} {
		tk_messageBox -title "Kill PGC Playback" -icon warning -parent . -message \
				"WARNING:\n$pgclabel:\nLink to Cell, PG or PTT command with\nhighlight button not changed to GoTo."
	}
	if {$pgcvar_arg == ""} {
		refresh_current_pgc
	}

	if {$type == "menu" && $numcells > 0 && $oktoblank} {
		if {$pgcvar_arg == ""} {
			set txt "You killed a Menu PGC playback."
			set txt "$txt\n"
			set txt "$txt\nDo you want to remove also the"
			set txt "$txt\nPrograms and Cells from the PGC?"
			set txt "$txt\nThis will allow to easily strip the"
			set txt "$txt\ncell with VobBlanker."
			set r [tk_messageBox -title "Kill PGC Playback" -icon question -message $txt \
					-parent . -type yesno -default yes]
		} else {
			set r "yes"
		}
		if {$r == "yes"} {
			set ::menus($index,numprograms) 0
			set ::menus($index,programmap) ""
			set ::menus($index,numcells) 0
			set ::menus($index,cellpbtable) ""
			set ::menus($index,cellpostable) ""
			::main::init_tree_menu_no_move
		}
	}
	::main::dvd_modified true false
	return 1
}

# check if it is safe to kill playback on the PGC
# Returns 0 if not safe, 1 if safe, or -1 for a dummy PGC (not needed)
# It may not be safe to kill playback when there are no post commands, but this is NOT checked!
proc check_safe_kill_pb {type vts lu pgc} {
	if {$type == "menu"} {
		set index "$vts,$lu,$pgc"
		set pgcvar ::menus($index)
		set numcells $::menus($index,numcells)
		set cellpbtable $::menus($index,cellpbtable)
	} else {
		set index "$vts,$pgc"
		set pgcvar ::pgcs($index)
		set numcells $::pgcs($index,numcells)
		set cellpbtable $::pgcs($index,cellpbtable)
	}
	if {$numcells == 0} {
		return -1	;# dummy PGC!
	}
	set cellcmds [string toupper [get_type_cmds cell $pgcvar]]
	set numcellcmds [llength $cellcmds]
	if {$numcellcmds == 0} {
		return 1	;# no cell commands
	}
	set postcmds [string toupper [get_type_cmds post $pgcvar]]
	if {[llength $postcmds] == 1} {
		set onlyonepostcmd [lindex $postcmds 0]
	} else {
		set onlyonepostcmd ""
	}
	# check the cell commands
	for {set cell 1} {$cell <= $numcells} {incr cell} {
		set cellcmdnum [get_int cellpbtable [expr {(($cell-1)*24)+3}] 1]
		if {$cellcmdnum == 0} continue	;# no cell command for this cell
		if {$cellcmdnum > $numcellcmds} continue	;# no cell command defined for this cell command number
		set cellcmd [lindex $cellcmds [expr {$cellcmdnum-1}]]
		if {$cellcmd == $onlyonepostcmd} continue	;# this cell command is the same as the only post command
		if {[regexp {20 .1 .. .. .. .. .. 10} $cellcmd]} {
			return 0	;# RSM cell-command (not checked by highlight_jump_cmd)
		}
		set target [::main::highlight_jump_cmd $type $index \[${cellcmd}\] -1 false false]
		if {$target == ""} continue	;# cell command jumps inside the same PGC or is not a Jump/Call/Link cmd
		set targettype [lindex $target 0]
		if {$targettype != $type} {
			return 0	;# jump to another domain
		}
		if {$targettype == "menu"} {
			set targetindex [lindex $target 1],$lu,[lindex $target 3]
		} else {
			set targetindex [lindex $target 1],[lindex $target 3]
		}
		if {$targetindex != $index} {
			return 0	;# jump to another PGC or domain
		}
	}
	return 1
}

#############################################################################
## Enable all operations for the current PGC
proc enable_all_operations {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set type  [lindex $::editor_cmds_from 0]
	set index [lindex $::editor_cmds_from 1]

	# clear PUOs
	set_int ::${type}s($index,header) 8 0 4

	# if it is a Title PGC, clear also the PUOs in the TT_SRPT table
	if {$type == "pgc" && [info exists ::${type}s($index,Title)]} {
		set data $::raw_sectors(0,VMG_TT_SRPT)
		set title [subst \$::${type}s($index,Title)]
		set numtitles [get_int data 0 2]
		if {$numtitles >= $title} {
			set t_idx [expr {8+(($title-1)*12)}]
			set typebyte [expr {[get_int data $t_idx 1] & 0xFC}]
			set_int ::raw_sectors(0,VMG_TT_SRPT) $t_idx $typebyte 1
		}
	}

	# enable navigation buttons
	set txt ""
	if {[subst \$::${type}s($index,numprograms)] > 0} {
		set pgcnum [lindex [split $index ","] end]
		if {[get_int ::${type}s($index,header) 0x9C 2] == 0} {
			set cmds [subst \$::${type}s($index)]
			if {[num_type_cmds "post" cmds] > 0} {
				set_int ::${type}s($index,header) 0x9C $pgcnum 2
			}
		}
		if {[get_int ::${type}s($index,header) 0x9E 2] == 0} {
			set_int ::${type}s($index,header) 0x9E $pgcnum 2
		}
		if {$type == "menu" && [get_int ::${type}s($index,header) 0xA0 2] == 0} {
			set_int ::${type}s($index,header) 0xA0 65535 2
			set txt " and Prev, Next and GoUp"
		} else {
			set txt " and Prev and Next"
		}
		set txt "$txt navigation buttons enabled"
	}

	show_msg "All Prohibited User Operations cleared$txt."
	::main::dvd_modified true false
	return
}

#############################################################################
## Completely remove the parental managment from a DVD
proc remove_parental_managment {} {
	set utiltitle "Remove Parental Managment"
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	if {!$::config(expert_mode)} {
		set txt "This utility will remove the parental managment:"
		set txt "$txt\ndelete the VMG_PTL_MAIT table, reset all Parental ID"
		set txt "$txt\nmasks in the VMG_TT_SRPT and the VTS_PGCITI tables,"
		set txt "$txt\nand change the SetTmpPML commands to GOTOs.\n"
		set txt "$txt\nThe DVD will therefore ignore the parental level"
		set txt "$txt\nof the player.\n"
		set txt "$txt\nAre you sure ?"
		set r [tk_messageBox -title $utiltitle -icon warning -message $txt \
				-parent . -type yesno -default no]
		if {$r == "no"} return
	}

	# remove the VMG_PTL_MAIT table
	if {$::raw_sectors(0,VMG_PTL_MAIT) != ""} {
		set ::raw_sectors(0,VMG_PTL_MAIT) ""
		set log "The VMG_PTL_MAIT table has been erased."
	} else {
		set log "The VMG_PTL_MAIT table was not defined."
	}

	# Reset all PM Masks for all titles in VMG_TT_SRPT to 0.
	set numtitles [get_int ::raw_sectors(0,VMG_TT_SRPT) 0 2]
	for {set t 1} {$t <= $numtitles} {incr t} {
		set_int ::raw_sectors(0,VMG_TT_SRPT) [expr {$t*12}] 0 2
	}
	set log "${log}\nParental ID masks for $numtitles Titles cleared in VMG_TT_SRPT table."


	# Reset the Category Parental ID mask for all PGCs in all VTS_PGCI to 0.
	set numchanges 0
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		set numpgcs [get_int ::raw_sectors($vts,VTS_PGCI) 0 2]
		for {set pgc 1} {$pgc <= $numpgcs} {incr pgc} {
			set idx [expr {2+($pgc*8)}]
			set_int ::raw_sectors($vts,VTS_PGCI) $idx 0 2
		}
		incr numchanges $numpgcs
	}
	set log "${log}\nParental ID masks of $numchanges PGCs cleared in $::pgcs(numvts) VTS_PGCITI tables."

	# Change all SetTmpPML to Goto (to same line).
	set numchanges 0
	for {set lu 1} {$lu <= $::menus(0,numlus)} {incr lu} {
		for {set pgc 1} {$pgc <= $::menus(0,$lu,numpgcs)} {incr pgc} {
			set idxs [lsearch -regexp -all $::menus(0,$lu,$pgc) {00 .3 .. .. .. .. .. ..}]	;# SetTmpPML
			foreach idx $idxs {
				set scmd [lindex $::menus(0,$lu,$pgc) $idx]
				set gcmd [string range $scmd 0 3]1[string range $scmd 5 17]00[string range $scmd 20 end]
				lset ::menus(0,$lu,$pgc) $idx $gcmd
				incr numchanges
			}
		}
	}
	switch $numchanges {
		0 {set log "${log}\nNo SetTmpPML command found in the VMGM."}
		1 {set log "${log}\nOne SetTmpPML command changed to GoTo in the VMGM."}
		default {set log "${log}\n$numchanges SetTmpPML commands changed to GoTo in the VMGM."}
	}

	# Change all instances of SPRM(13) by constant 15: left to user.

	# Done. Refresh interface
	::main::dvd_modified true false
	::main::display_pgc pgc 0 0

	set txt "Done:\n$log\n"
	set txt "${txt}You will never be prompted for a password anymore!\n\n"
	set txt "${txt}To be sure that the Parental level is never used in the DVD,\n"
	set txt "${txt}you should search for any occurence of sprm(13:Parental level)\n"
	set txt "${txt}and change the sprm 13 in the command by the constant value 15.\n"
	set txt "${txt}For example, change \"Set gprm(4) =(mov) sprm(13:Parental level)\"\n"
	set txt "${txt}to \"Set gprm(4) =(mov) 15\".\n"
	set txt "${txt}\nDo you want to search for \"sprm(13:\" now?  (recommended)"

	set r [tk_messageBox -title $utiltitle -icon info -message $txt \
			-parent . -type yesno -default yes]
	if {$r == "yes"} {
		::utils::find_results {sprm(13:} false false
	}
	return
}

#############################################################################
## Find Unused GPRMs
proc find_unused_gprms {quiet {return_usage_count false}} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}

	show_msg "Searching for GPRMs..."

	for {set i 0} {$i < 16} {incr i} {
		set used($i) 0
	}

	# FP PGC
	set r [get_used_gprms ::pgcs(0,0)]
	for {set i 0} {$i < 16} {incr i} {
		incr used($i) [lindex $r $i]
	}

	# Titles
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		for {set ttn 1} {$ttn <= $::pgcs($vts,numpgcs)} {incr ttn} {
			set r [get_used_gprms ::pgcs($vts,$ttn)]
			for {set i 0} {$i < 16} {incr i} {
				incr used($i) [lindex $r $i]
			}
		}
	}

	# Menus
	for {set vts 0} {$vts <= $::menus(numvts)} {incr vts} {
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				set r [get_used_gprms ::menus($vts,$lu,$pgc)]
				for {set i 0} {$i < 16} {incr i} {
					incr used($i) [lindex $r $i]
				}
			}
		}
	}

	# Menu buttons
	for {set vts 0} {$vts <= $::menus(numvts)} {incr vts} {
		set butcmds {}
		if {[info exists ::buttons::butsinfo(menu,$vts,errorcode)] && \
				$::buttons::butsinfo(menu,$vts,errorcode) == 0} {
			for {set vid 1} {$vid <= $::buttons::butsinfo(menu,$vts,numvobids)} {incr vid} {
				for {set cid 1} {$cid <= $::buttons::butsinfo(menu,$vts,$vid,numcellids)} {incr cid} {
					if {[info exists ::buttons::butsinfo(menu,$vts,$vid,$cid,numbuts)] && \
								$::buttons::butsinfo(menu,$vts,$vid,$cid,numbuts) > 0} {
						for {set group 1} {$group <= $::buttons::butsinfo(menu,$vts,$vid,$cid,numgroups)} {incr group} {
							for {set but 1} {$but <= $::buttons::butsinfo(menu,$vts,$vid,$cid,numbuts)} {incr but} {
								lappend butcmds $::buttons::butsinfo(menu,$vts,$vid,$cid,$group,$but,command)
							}
						}
					}
				}
			}
		}
		set r [get_used_gprms butcmds]
		for {set i 0} {$i < 16} {incr i} {
			incr used($i) [lindex $r $i]
		}
	}
	unset butcmds

	# BOVs
	for {set vts 0} {$vts <= $::pgcs(numvts)} {incr vts} {
		set butcmds {}
		if {[info exists ::buttons::butsinfo(pgc,$vts,errorcode)] && \
				$::buttons::butsinfo(pgc,$vts,errorcode) == 0} {
			for {set vid 1} {$vid <= $::buttons::butsinfo(pgc,$vts,numvobids)} {incr vid} {
				for {set cid 1} {$cid <= $::buttons::butsinfo(pgc,$vts,$vid,numcellids)} {incr cid} {
					if {[info exists ::buttons::butsinfo(pgc,$vts,$vid,$cid,numbuts)] && \
								$::buttons::butsinfo(pgc,$vts,$vid,$cid,numbuts) > 0} {
						for {set group 1} {$group <= $::buttons::butsinfo(pgc,$vts,$vid,$cid,numgroups)} {incr group} {
							for {set but 1} {$but <= $::buttons::butsinfo(pgc,$vts,$vid,$cid,numbuts)} {incr but} {
								lappend butcmds $::buttons::butsinfo(pgc,$vts,$vid,$cid,$group,$but,command)
							}
						}
					}
				}
			}
		}
		set r [get_used_gprms butcmds]
		for {set i 0} {$i < 16} {incr i} {
			incr used($i) [lindex $r $i]
		}
	}
	unset butcmds

	set txt ""
	set lst ""
	set usedtxt ""
	for {set i 0} {$i < 16} {incr i} {
		if {$used($i) == 0} {
			lappend lst $i
			set txt "$txt$i, "
		} else {
			set usedtxt "$usedtxt\ngprm($i) : $used($i)"
		}
	}
	if {$txt==""} {
		set txt "All GPRMs are used!"
	} else {
		set txt "Unused GPRM numbers:\n[string trim $txt ", "]."
	}
	if {!$quiet} {
		show_msg $txt
		set txt "$txt\n\nUsage count\n(including menubuttons and BOVs):$usedtxt"
		tk_messageBox -title "Find Unused GPRMs" -icon info -message $txt -parent .
	}
	if {$return_usage_count} {
		set lst ""
		for {set i 0} {$i < 16} {incr i} {
			lappend lst $used($i)
		}
	}
	return $lst
}
proc get_used_gprms {cmds_upvar} {
	upvar $cmds_upvar cmds
	for {set i 0} {$i < 16} {incr i} {
		set used($i) 0
	}
	foreach cmd $cmds {
		if {[string range $cmd 0 0] == "*"} continue
		set d [::disasm::disasm $cmd]
		set idx [string first "gprm(" $d]
		while {$idx >= 0} {
			incr idx 5
			scan [string range $d $idx end] "%d" g
			incr used($g)
			set idx [string first "gprm(" $d $idx]
		}
	}
	set allused ""
	for {set i 0} {$i < 16} {incr i} {
		lappend allused $used($i)
	}
	return $allused
}

#############################################################################
## Search for all used GPRMs in the DVD and insert the commands to reset them
## after the current line
proc reset_used_gprms {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	if {!$::config(expert_mode)} {
		set txt "This macro will search for all used GPRMs"
		set txt "$txt\nand insert the commands to reset them to 0"
		set txt "$txt\nAFTER the current selection."
		set txt "$txt\nThe GOTOs lines are renumbered if needed."
		set txt "$txt\n"
		set txt "$txt\nContinue ?"
		set r [tk_messageBox -title "Reset Used GPRMs" -icon question -message $txt \
				-parent . -type yesno -default yes]
		if {$r == "no"} return
	}

	# find all used GPRMs
	set unusedgprms [find_unused_gprms true]
	set usedcmds ""
	for {set u 0} {$u < 16} {incr u} {
		if {[lsearch $unusedgprms $u] == -1} {
			lappend usedcmds "71 00 00 [format "%02X" $u] 00 00 00 00"
		}
	}
	set numused [llength $usedcmds]
	if {$numused == 0} return

	# split pre/post/cell commands
	set idxs [::main::get_selection_indexes]
	set idx1 [lindex $idxs 0]
	set idx2 [lindex $idxs 1]
	if {$idx1 == "-"} {
		bell
		return ""
	}
	set cmds [subst \$::[lindex $::editor_cmds_from 0]s([lindex $::editor_cmds_from 1])]
	set line 0
	set lines(pre)  ""
	set lines(post) ""
	set lines(cell) ""
	foreach c $cmds {
		if {[string range $c 0 0] == "*"} {
			set type [lindex $c 1]
			set numcmds($type) 0
		} else {
			incr numcmds($type)
			lappend lines($type) $c
		}
		if {$line == $idx2} {
			set instype $type
			set insln $numcmds($type)
		}
		incr line
	}

	set cmds $lines($instype)

	# renumber GOTO lines
	for {set idx 0} {$idx < [llength $cmds]} {incr idx} {
		set cmd [lindex $cmds $idx]
		set cmdchk "[string range $cmd 0 1][string range $cmd 4 4]"
		if {$cmdchk == "001" || $cmdchk == "003"} {
			set ln [x2d [lindex $cmd end]]
			if {$ln > $insln} {
				incr ln $numused
				lset cmd end [format "%02X" $ln]
				lset cmds $idx $cmd
			}
		}
	}

	# rebuild full PGC commands
	set cmds1 [lrange $cmds 0 [expr {$insln-1}]]
	set cmds2 [lrange $cmds $insln end]
	set lines($instype) [concat $cmds1 $usedcmds $cmds2]
	set ::[lindex $::editor_cmds_from 0]s([lindex $::editor_cmds_from 1]) \
			[concat {{* pre}} $lines(pre) {{* post}} $lines(post) {{* cell}} $lines(cell)]

	refresh_current_pgc
	::main::dvd_modified true false
}

#############################################################################
## Show current PGC as text
proc show_current_pgc_as_text {} {
	if {$::scratchbook_active} {
		set title "Scratchbook"
		set txt ""
	} else {
		if {! [info exists ::dvddir]} {
			bell
			return ""
		}
		set title "[$::widget(main_cmd_lab) cget -text] -[$::widget(main_cmd_lab2) cget -text]"
		set txt "$title\n\n"
	}
	foreach i $::editor_cmds_list {
		if {$::config(dump_hex_codes)} {
			set txt "$txt$i\n"
		} else {
			if {[string range $i 0 0] == "*"} {
				set txt "$txt$i\n"
			} else {
				set txt "$txt[string range $i 25 end]\n"
			}
		}
	}
	if {[info exists ::dvddir]} {
		set type [lindex $::editor_cmds_from 0]
		set index [lindex $::editor_cmds_from 1]

		if {!$::scratchbook_active} {
			set vts [lindex [split $index ","] 0]
			if {$type == "menu"} {
				set txt "$txt\n[::menubuttons::dump_pgc_menu_buttons $vts \
						$::menus($index,numcells) $::menus($index,cellpostable) "" $::config(dump_hex_codes)]"
			} elseif {$vts > 0} {
				set txt "$txt\n[::menubuttons::dump_pgc_BOVs $vts \
						$::pgcs($index,numcells) $::pgcs($index,cellpostable) "" $::config(dump_hex_codes)]"
			}
			set pgcheader    [subst \$::${type}s($index,header)]
			set numprograms  [subst \$::${type}s($index,numprograms)]
			set programmap   [subst \$::${type}s($index,programmap)]
			set numcells     [subst \$::${type}s($index,numcells)]
			set cellpbtable  [subst \$::${type}s($index,cellpbtable)]
			set cellpostable [subst \$::${type}s($index,cellpostable)]
			if {$type == "pgc" && [info exists ::${type}s($index,pttlist)]} {
				set pttlist [subst \$::${type}s($index,pttlist)]
			} else {
				set pttlist ""
			}
			set vtsnum [lindex [split $index ","] 0]

			set txt "$txt\n\n[::pgced::dump_pgc \
					$type $index $pgcheader $numprograms $programmap \
					$numcells $cellpbtable $cellpostable $pttlist]"
		}
	}

	::main::showtext .curpgc $title $txt
}

proc explore_dvd {args} {
	if {$::tcl_platform(platform) != "windows"} return
	if {! [info exists ::dvddir]} {
		bell
		return
	}

	if {$args == "parent"} {
		exec -- explorer.exe [file nativename [file dirname $::dvddir]] &
	} else {
		exec -- explorer.exe [file nativename $::dvddir] &
	}
}

proc jump2pgc {} {
	set utiltitle "Jump to PGC upon DVD insert"
	if {![info exists ::dvddir] || $::scratchbook_active} {bell;return}

	set cur_type [lindex $::editor_cmds_from 0]
	set cur_index [lindex $::editor_cmds_from 1]
	set cur_vts   [lindex [split $cur_index ","] 0]
	if {$cur_type == "menu"} {
		set cur_lu    [lindex [split $cur_index ","] 1]
		set cur_pgc   [lindex [split $cur_index ","] 2]
	} else {
		set cur_pgc   [lindex [split $cur_index ","] 1]
	}

	set txt "This macro will modify the VM commands of several PGCs"
	set txt "$txt\nso that the current PGC will be played upon DVD insert."
	set txt "$txt\n\nIt is highly recommended to do a backup of your IFOs"
	set txt "$txt\nbefore using this macro!  (Use File -> Incremental Backup.)"

	set txt "$txt\n\nTo use this macro, you have to trace from the First-Play PGC"
	set txt "$txt\nuntil you reach the target PGC.  The cursor must be on the first"
	set txt "$txt\nor second line of the target PGC, ie. on the \"*** pre commands\""
	set txt "$txt\nlabel or on the first pre command (or on the \"*** post commands\""
	set txt "$txt\nlabel if there are no pre commands in the PGC.)"
	set txt "$txt\n\nHINT: Set a pre/post breakpoint on your target PGC, then restart"
	set txt "$txt\nthe trace with the Load button, and use the \"->>> Run\" button."
	set txt "$txt\nThe playback will automatically stop at the right location."
	if {$cur_vts != 0} {
		set txt "$txt\n\nIf you want to select default audio and/or subtitle, feel free"
		set txt "$txt\nto continue the trace, go to the audio/subtitle menus, select"
		set txt "$txt\nthe streams you want to use, and return to the target PGC.\n"
	} else {
		set txt "$txt\n\nNOTE: It is not possible to set the default audio and subpic"
		set txt "$txt\nstreams, because the target PGC is in the VMGM domain!\n"
	}
	set ok true

	# verify PGC is not a dummy
	if {[subst \$::${cur_type}s($cur_index,numprograms)] == 0} {
		set ok false
		set txt "$txt\n\nERROR: the current PGC must have video contents."		
		set txt "$txt\nPlease select a playable menu or title PGC."		
		set txt "$txt\n(The PGC may have menu buttons or BOVs.)"		
	}

	# in Title domain, the PGC must be a Title entry
	if {$ok && $cur_type == "pgc"} {
		if {![info exists ::pgcs($cur_index,Title)] || $::pgcs($cur_index,Title)==""} {
			set ok false
			set txt "$txt\n\nERROR: To jump to a PGC in the title domain, you must select a"		
			set txt "$txt\ntitle entry (ie. a PGC with a Title number in the PGC Selector)."		
		}
	}

	# the 1st or 2nd line must be selected.
	set oktobreak false
	if {$ok} {
		set oktobreak true
		set idxs [::main::get_selection_indexes]
		set idx1 [lindex $idxs 0]
		set idx2 [lindex $idxs 1]
		if {$idx1 != $idx2} {set ok false}
		if {$idx1 > 1} {set ok false}
		if {! $ok} {
			set txt "$txt\n\nERROR: Wrong line selected in a valid PGC."
		}
	}

	# verify that trace mode is enabled
	set trace_opened true
	if {! [winfo exists .mf.cf.trace] || ! [winfo ismapped .mf.cf.trace]} {
		if {$ok} {
			set txt "$txt\n\nERROR: Trace mode not enabled!"
			set txt "$txt\nPgcEdit will now start the Trace mode."
		} else {
			set txt "$txt\n\nPgcEdit will now start the Trace mode."
		}
		set trace_opened false
		set ok false
	}
	
	# verify that the current cursor position matches the internal trace variables
	# (This is a basic check to ensure the PGC has been reached via the trace function)
	if {$ok} {
		if {$::tr::current(PGC)!=$cur_pgc} {
			set ok false
		}
		if {$ok} {
			if {$cur_type == "menu"} {
				if {$::tr::current(LU)!=$cur_lu} {
					set ok false
				}
			} else {
				if {([info exists ::pgcs($cur_index,Title)] && $::tr::current(Title)!=$::pgcs($cur_index,Title)) \
							|| $::tr::current(TTN)!=$::pgcs($cur_index,TTN)} {
					set ok false
				}
			}
		}
		if {! $ok} {
			set txt "$txt\n\nERROR: You must go to the target PGC using the trace."
			set txt "$txt\nDo not select a PGC manually in the PGC Selector pane,"
			set txt "$txt\nand do not select a command manually in the main pane!"
			set txt "$txt\nYou should use only the trace navigation buttons (and"
			set txt "$txt\nselect the right DVD menu buttons if needed.)"
		}
	}

	# verify that the PGC is entered straight at cell 1, program 1 (chapter 1).  Warning only.
	if {$ok} {
		if {$::tr::current(Cell)!=1 || $::tr::current(PG)!=1} {
			set ok false
		} elseif {$cur_type == "pgc" && $::tr::current(PTT)!=1} {
			set ok false
		}
		if {! $ok} {
			set txt "$txt\n\nWARNING: The PGC was not entered in cell 1, program 1."
			if {$cur_type == "pgc" && $::tr::current(PTT)!=1} {
				set txt "$txt\nIf you have selected a chapter in the chapter menu to"
				set txt "$txt\njump to the title, please start the trace again, but"
				set txt "$txt\ndo not use the chapter menu (or select chapter 1.)"
			}
			set txt "$txt\nYou may try to continue anyway, but the macro may fail,"
			set txt "$txt\nor the beginning of the PGC (cell 1) may be played, even"
			set txt "$txt\nif it is not what you want."
			set ok true
		}
	}

	# display tests results.  Open trace mode if needed.  Exit if error.
	if {$ok} {
		set txt "$txt\n\nIf everything is correct, select OK to continue..."
		set r [tk_messageBox -title $utiltitle -icon question -message $txt -parent . -type okcancel -default ok]
		if {$r == "cancel"} {
			set ok false
		}
	} else {
		set txt "$txt\n\nTrace to the target PGC now, and call this macro again..."
		tk_messageBox -title $utiltitle -icon warning -message $txt -parent .
		if {$oktobreak} {
			if {! [info exists ::tr::breakpoints([lindex $::editor_cmds_from 0],[lindex $::editor_cmds_from 1],prepost)]} {
				set txt "The currently selected PGC is valid, but has no pre/post breakpoint."
				set txt "$txt\nPgcEdit can set a pre/post breakpoint to this PGC now, if you want."
				set txt "$txt\n(Recommended if the PGC is really the target you want to jump to.)"
				set r [tk_messageBox -title $utiltitle -icon question -message $txt -parent . -type yesno -default yes]
				if {$r == "yes"} {
					set ::tr::breakpoints([lindex $::editor_cmds_from 0],[lindex $::editor_cmds_from 1],prepost) 1
					::main::highlight_trace_breakpoints
				}
			}
		}
		if {!$trace_opened} {
			::tr::dvd_trace
		}
	}
	if {! $ok} {
		return
	}
	
	###### Everything is correctly setupped now.  Do the job!
	
	# search a suitable GPRM to use
	set gprmusage [find_unused_gprms true true]
	set freegprms ""
	for {set i 0} {$i < 16} {incr i} {
		if {[lindex $gprmusage $i] == 0} {
			lappend freegprms $i
		}
	}
	if {$freegprms != ""} {
		set safe_gprm true
		set usecount 0
		set gprm [lindex $freegprms [expr {[llength $freegprms]-1}]]
	} else {
		set safe_gprm false
		# search for a GPRM that has not been set yet.
		# If there are several of them, choose the one with the lower usage count possible.
		set gprm -1
		set usecount 9999
		for {set i 15} {$i >= 0} {incr i -1} {
			if {$::tr::gprm_changed($i)==0} {
				if {[lindex $gprmusage $i] < $usecount} {
					set usecount [lindex $gprmusage $i]
					set gprm $i
				}
			}
		}
		if {$gprm == -1} {
			# pick an used GPRM with the lower usage count possible.
			set gprm -1
			set usecount 9999
			for {set i 15} {$i >= 0} {incr i -1} {
				if {[lindex $gprmusage $i] < $usecount} {
					set usecount [lindex $gprmusage $i]
					set gprm $i
				}
			}
		}
	}
	set hexgprm [format "%02X" $gprm]
	show_msg "Macro running... PLEASE WAIT!"

	# find the PGC to jump to.
	# If it is a title PGC, it's the same as the target PGC.  Else, it must be an entry PGC.
	# If there are no entry PGCs in the domain, we must create one.
	set jump_type 0
	if {$cur_type == "menu" && $cur_vts != 0} {
		if {[expr $::menus($cur_index,mask) & 0x0F] == 0} {
			for {set pgc 1} {$pgc <= $::menus($cur_vts,$cur_lu,numpgcs)} {incr pgc} {
				if {[expr $::menus($cur_vts,$cur_lu,$pgc,mask) & 0x0F] != 0} {
					set jump_index $cur_vts,$cur_lu,$pgc
					set jump_pgc   $pgc
					set jump_type  $::menus($cur_vts,$cur_lu,$pgc,mask)
					break
				}
			}
			if {$jump_type == 0} {
				set newtype 6
				set newtypetext "AngleM"
				set jump_index $cur_index
				set jump_pgc   $cur_pgc
				set jump_type  $newtype				
				set ::menus($cur_index,mask) $newtype
				set ::menus($cur_index,masktext) $newtypetext
				::main::init_tree_menu_no_move
				update idletasks
				set txt "WARNING: It is not possible to jump directly to your target PGC."
				set txt "$txt\nTherefore, PgcEdit has changed the type of the target PGC to \"$newtypetext\"."
				set txt "$txt\nYou may need to prohibit the use of the ${newtypetext}enu button."
				tk_messageBox -title $utiltitle -icon warning -message $txt -parent .
			}
		} else {
			set jump_index $cur_index
			set jump_pgc   $cur_pgc
			set jump_type  $::menus($cur_index,mask)
		}
	} else {
		set jump_index $cur_index
		set jump_pgc   $cur_pgc
		set jump_type  -1
	}

	# If in the menu domain, remember the last title that has been entered (resume point).
	set rsm_warning false
	if {$cur_type == "menu" && [info exists ::tr::rsm_vts] && $::tr::rsm_vts >= 1} {
		set rsm_warning true
		if {( $::tr::rsm_vts == $cur_vts || $cur_vts == 0 ) && \
			! [catch {set rsm_title $::pgcs($::tr::rsm_vts,$::tr::rsm_pgc,Title)}]} {
			set rsm_index $::tr::rsm_vts,$::tr::rsm_pgc
			set rsm_vts   $::tr::rsm_vts
			set rsm_pgc   $::tr::rsm_pgc
			set rsm_type  -1
		} else {
			set rsm_index ""
		}
	} else {
		set rsm_index ""
	}

	# in menu domain, if there are several LUs, warn the user and ask what to do
	if {$cur_type == "menu" && $::menus($cur_vts,numlus)>1} {
		set identical true
		for {set lu 2} {$lu <= $::menus($cur_vts,numlus)} {incr lu} {
			if {$::menus($cur_vts,1,$cur_pgc)!=$::menus($cur_vts,$lu,$cur_pgc)} {
				set identical false
			}
			if {$::menus($cur_vts,1,$jump_pgc)!=$::menus($cur_vts,$lu,$jump_pgc)} {
				set identical false
			}
		}
		set txt "There are several Language Units in your target menu domain."
		set txt "$txt\nThe macro should be applied to all LUs of the domain, or the"
		set txt "$txt\njump will not work if you use another default menu language"
		set txt "$txt\nconfiguration in your player.\n"
		if {$::menus($cur_vts,lus_are_identical) && $identical} {			
			set txt "$txt\nAs the commands of the PGCs to be modified are identical in"
			set txt "$txt\nall LUs, the macro should process the PGCs in every LU."
			set txt "$txt\n\nApply the macro to all LUs ?  (Recommended)"
			set icon question
			set default yes
		} elseif {!$identical} {
			set txt "$txt\nWARNING: The commands in the PGCs to be processed are NOT"
			set txt "$txt\nidentical.  Applying the macro on all LUs anyway may lead"
			set txt "$txt\nto navigation errors.  If you do so, you will need to test"
			set txt "$txt\ncarefully the DVD in every menu language configuration."
		} else {
			set txt "$txt\nWARNING: When loading the DVD, PgcEdit has detected some"
			set txt "$txt\ndifferences in the commands of the LUs in this titleset."
			set txt "$txt\nHowever, the commands in the PGCs to be processed are"
			set txt "$txt\nidentical.  You may try to apply the macro on all LUs."
			set txt "$txt\nHowever, you will need to test carefully the DVD in every"
			set txt "$txt\nmenu language configuration."
		}
		if {! $::menus($cur_vts,lus_are_identical) || ! $identical} {			
			set txt "$txt\n\nIf you prefer, you may also run the macro again (including"
			set txt "$txt\nthe trace) in every LU language configyration."
			set txt "$txt\nYou will need to change the Preferred Menu Language code"
			set txt "$txt\nin the virtual player setup menu, and restart the trace and"
			set txt "$txt\nthe macro, for each language present in the LUs."
			set txt "$txt\n\nYou may also apply the macro on the current LU only, and"
			set txt "$txt\nthen, use the 'Copy LU cmds to other LUs' function to clone"
			set txt "$txt\nthe current LU behaviour in the other LUs, but you will lose"
			set txt "$txt\nthe LU differences, and that may lead to navigation problems."
			set txt "$txt\n\nApply the macro to all LUs ?  (Dangerous!)"
			set icon warning
			set default no
		}
		set r [tk_messageBox -title $utiltitle -icon $icon -message $txt -parent . \
				-type yesnocancel -default $default]
		if {$r == "cancel"} return
		if {$r == "no"} {
			set first_lu $cur_lu
			set last_lu  $cur_lu
		} else {
			set first_lu 1
			set last_lu  $::menus($cur_vts,numlus)
		}
	} elseif {$cur_type == "menu"} {
		set first_lu 1
		set last_lu  1
	} else {
		set first_lu -1
		set last_lu  -1
	}

	# Snapshot the current state (generate the Set GPRM, SetSTN and SetHL_BTN commands)
	set set_commands [::tr::copy_regs_as_cmds changed none]
	if {$cur_vts == 0} {
		# SetSTN is illegal in VMGM
		set idx [lsearch -regexp $set_commands {51 .. .. .. .. .. .. ..}]
		if {$idx != -1} {
			set set_commands [lreplace $set_commands $idx $idx]
		}
	}

	# Now, continue the trace to verify that the PGC is correctly played
	show_msg "Macro running... PLEASE WAIT!"
	set stopline [expr [num_type_cmds pre ::${cur_type}s($cur_index)]+1]
	set curline [lindex [::main::get_selection_indexes] 0]
	while {$curline != $stopline} {
		# advance one step in trace
		::tr::display_current_line
		::tr::exec_current_line
		::tr::reset_log_highlight
		if {$cur_type != [lindex $::editor_cmds_from 0] || \
				$cur_index != [lindex $::editor_cmds_from 1]} {
			set txt "The pre-commands of your target PGC jumped outside the PGC"
			set txt "$txt\nbefore the playback has been reached."
			set txt "$txt\n\nPlease continue the trace until you reach the PGC again,"
			set txt "$txt\nand call this macro again when done."
			tk_messageBox -title $utiltitle -icon warning -message $txt -parent .
			return
		}
		set curline [lindex [::main::get_selection_indexes] 0]
	}

	# generate the commands in the target PGC
	set icmds {{* pre}}
	if {$safe_gprm} {
		# 1  if ( gprm(G) != 0 ) then { Goto line L }
		lappend icmds "00 B1 00 $hexgprm 00 00 00 [format {%02X} [expr [llength $set_commands]+3]]"
		# concat 'Set GPRN(changed) =(mov) value' lines.
		set icmds [concat $icmds $set_commands]
		# L-1  Set gprm(G) =(mov) 65278
		lappend icmds "71 00 00 $hexgprm FE FE 00 00"
	} else {
		# 1  if ( gprm(G) != 65278 ) then { Goto line L } 
		lappend icmds "00 B1 00 $hexgprm FE FE 00 [format {%02X} [expr [llength $set_commands]+3]]"
		# 2  Set gprm(G) =(mov) 0
		lappend icmds "71 00 00 $hexgprm 00 00 00 00"
		# concat 'Set GPRN(changed) =(mov) value' lines.
		set icmds [concat $icmds $set_commands]
	}
	# L  NOP
	lappend icmds "00 00 00 00 00 00 00 00"
	set numnewcmds [expr [llength $icmds] -1]

	for {set lu $first_lu} {$lu <= $last_lu} {incr lu} {
		if {$lu == -1} {
			set index $cur_vts,$cur_pgc
		} else {
			set index $cur_vts,$lu,$cur_pgc
		}
		set origcmds [lreplace [subst \$::${cur_type}s($index)] 0 0]
		# renumber GOTO lines
		for {set idx 0} {$idx < [llength $origcmds]} {incr idx} {
			set cmd [lindex $origcmds $idx]
			if {[string range $cmd 0 0] == "*"} break
			set cmdchk "[string range $cmd 0 1][string range $cmd 4 4]"
			if {$cmdchk == "001" || $cmdchk == "003"} {
				set ln [expr [x2d [lindex $cmd end]] + $numnewcmds]
				lset cmd end [format "%02X" $ln]
				lset origcmds $idx $cmd
			}
		}
		set icmds [concat $icmds $origcmds]
		set ::${cur_type}s($index) $icmds
	}

	# If a suitable resume point is alerady defined, generate the commands in the resume title
	if {$rsm_index != ""} {
		set icmds {{* pre}}
		if {$safe_gprm} {
			# 1  if ( gprm(G) != 0 ) then { Goto line 3 }
			lappend icmds "00 B1 00 $hexgprm 00 00 00 03"
		} else {
			# 1  if ( gprm(G) != 65278 ) then { Goto line L } 
			lappend icmds "00 B1 00 $hexgprm FE FE 00 03"
		}
		if {$cur_vts == 0} {
			set pgchex [format %04X $jump_pgc]
			# 2  (CallSS) Call the VMGM PGC TT, resume cell 1
			lappend icmds "30 08 [string range $pgchex 0 1] [string range $pgchex 2 3] 01 C0 00 00"
		} else {
			set vtshex [format %02X $cur_vts]
			set jump_hex [format %X [expr $jump_type & 0x0F]]
			# 2  (CallSS) Call the VTSM T menu of the current VTS, resume cell 1 
			lappend icmds "30 08 00 00 $vtshex 8$jump_hex 00 00"
		}
		# 3  NOP
		lappend icmds "00 00 00 00 00 00 00 00"

		set numnewcmds 3
		set origcmds [lreplace $::pgcs($rsm_index) 0 0]
		# renumber GOTO lines
		for {set idx 0} {$idx < [llength $origcmds]} {incr idx} {
			set cmd [lindex $origcmds $idx]
			if {[string range $cmd 0 0] == "*"} break
			set cmdchk "[string range $cmd 0 1][string range $cmd 4 4]"
			if {$cmdchk == "001" || $cmdchk == "003"} {
				set ln [expr [x2d [lindex $cmd end]] + $numnewcmds]
				lset cmd end [format "%02X" $ln]
				lset origcmds $idx $cmd
			}
		}
		set icmds [concat $icmds $origcmds]
		set ::pgcs($rsm_index) $icmds
	}
	

	# generate the commands in the FP-PGC
	set icmds {{* pre}}
	if {$safe_gprm} {
		# 1  if ( gprm(G) != 0 ) then { Goto line 3 }
		lappend icmds "00 B1 00 $hexgprm 00 00 00 03"
	} else {
		# Check there are jumps to the FP-PGC: risk of nav errors!
		unset -nocomplain ::crossrefs_calculated
		unset -nocomplain ::crossrefs
		::utils::build_crossrefs -1
		if {$::crossrefs(pgcs,0,0,numrefs)>0} {
			for {set i 0} {$i < 16} {incr i} {
				# 1->16  if ( gprm(N) != 0 ) then { Goto line 4 }
				lappend icmds "00 B1 00 [format "%02X" $i] 00 00 00 13"
			}
			# 17  Set gprm(G) =(mov) 65278 
			lappend icmds "71 00 00 $hexgprm FE FE 00 00"
		} else {
			# 1  if ( gprm(G) != 0 ) then { Goto line 4 }
			lappend icmds "00 B1 00 $hexgprm 00 00 00 04"
			# 2  Set gprm(G) =(mov) 65278 
			lappend icmds "71 00 00 $hexgprm FE FE 00 00"
		}
	}
	if {$rsm_index != ""} {
		set titlehex [format %02X $rsm_title]
		# 2/3/18  (JumpTT) Jump to Title TT
		lappend icmds "30 02 00 00 00 $titlehex 00 00"
	} elseif {$cur_vts == 0} {
		set pgchex [format %04X $jump_pgc]
		# 2/3/18  (JumpSS) Jump to VMGM PGC TT 
		lappend icmds "30 06 [string range $pgchex 0 1] [string range $pgchex 2 3] 00 C0 00 00"
	} elseif {$cur_type == "pgc"} {
		set titlehex [format %02X $::pgcs($cur_index,Title)]
		# 2/3/18  (JumpTT) Jump to Title TT
		lappend icmds "30 02 00 00 00 $titlehex 00 00"
	} else {
		set vtshex [format %02X $cur_vts]
		set jump_hex [format %X [expr $jump_type & 0x0F]]
		# 2/3/18  (JumpSS) Jump to VTSM T, T menu (TTN 1) 
		lappend icmds "30 06 00 01 $vtshex 8$jump_hex 00 00"
	}
	# 3/4/19  NOP
	lappend icmds "00 00 00 00 00 00 00 00"
	set numnewcmds [expr [llength $icmds] -1]
	set origcmds [lreplace $::pgcs(0,0) 0 0]
	# renumber GOTO lines
	for {set idx 0} {$idx < [llength $origcmds]} {incr idx} {
		set cmd [lindex $origcmds $idx]
		if {[string range $cmd 0 0] == "*"} break
		set cmdchk "[string range $cmd 0 1][string range $cmd 4 4]"
		if {$cmdchk == "001" || $cmdchk == "003"} {
			set ln [expr [x2d [lindex $cmd end]] + $numnewcmds]
			lset cmd end [format "%02X" $ln]
			lset origcmds $idx $cmd
		}
	}
	set icmds [concat $icmds $origcmds]
	set ::pgcs(0,0) $icmds


	# generate the command in the intermediate jump target pgc, if needed
	if {$cur_type == "menu" && $cur_vts != 0 && $jump_index != $cur_index} { 
		set icmds {{* pre}}
		set pgchex [format %04X $cur_pgc]
		if {$safe_gprm} {
			# 1  if ( gprm(G) == 0 ) then { LinkPGCN PGC TT } 
			lappend icmds "20 A4 00 $hexgprm 00 00 [string range $pgchex 0 1] [string range $pgchex 2 3]"
		} else {
			# 1  if ( gprm(G) == 65278 ) then { LinkPGCN PGC TT } 
			lappend icmds "20 A4 00 $hexgprm FE FE [string range $pgchex 0 1] [string range $pgchex 2 3]"
		}		

		for {set lu $first_lu} {$lu <= $last_lu} {incr lu} {
			set index $cur_vts,$lu,$jump_pgc
			set origcmds [lreplace $::menus($index) 0 0]
			# renumber GOTO lines
			for {set idx 0} {$idx < [llength $origcmds]} {incr idx} {
				set cmd [lindex $origcmds $idx]
				if {[string range $cmd 0 0] == "*"} break
				set cmdchk "[string range $cmd 0 1][string range $cmd 4 4]"
				if {$cmdchk == "001" || $cmdchk == "003"} {
					set ln [expr [x2d [lindex $cmd end]] + 1]
					lset cmd end [format "%02X" $ln]
					lset origcmds $idx $cmd
				}
			}
			set icmds [concat $icmds $origcmds]
			set ::menus($index) $icmds
		}
	}

	# All done
	::main::dvd_modified true false
	::tr::reset_trace

	# Search for RSM commands in the menu domains.  If one is found, it's dangerous.
	# TODO: Check also for GoUpPGCN links = 255 in the menu domains (?)
	set found false
	if {$rsm_warning && $cur_type == "menu"} {
		set resumeflag -1
		for {set i 0} {$i < 16} {incr i} {
			set resumegprms($i) 0
		}
		for {set vts 0} {$vts <= $::pgcs(numvts)} {incr vts} {
			for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
				for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
					# search for RSM commands
					set idxs [lsearch -all -regexp $::menus($vts,$lu,$pgc) "20 .1 .. .. .. .. .. 10"]
					if {$idxs != ""} {
						foreach idx $idxs {
							set found true
							set cmd [lindex $::menus($vts,$lu,$pgc) $idx]
							# if RSM is associated with a condition, and the GPRM used in the
							# condition is already used in the target PGC, then that GPRM is
							# a verry good candidate for the resume flag.
							if {[string range $cmd 3 3]!= 0} {
								set tmp [x2d [string range $cmd 10 10]]
								if {[lsearch -regexp $set_commands "71 00 00 0[format %X $tmp] .. .. 00 00"] >= 0} {
									set resumeflag $tmp
									incr resumegprms($tmp) 5
									if {$vts == $cur_vts} {incr resumegprms($tmp) 2}
								}
							}
						}
						set last [lindex $idxs end]

						# If there is a GPRM in this PGC, search for conditional GOTOs, BREAK, LINKs...
						# If one is found BEFORE the last line with a RSM, the GPRM used in the condition
						# may also be the resume flag
						set idxs [lsearch -all -regexp $::menus($vts,$lu,$pgc) {[02]0 [2Aa][124567] 00 0. .. .. .. ..}]
						foreach idx $idxs {
							if {$idx < $last} {
								set found true
								set cmd [lindex $::menus($vts,$lu,$pgc) $idx]
								set tmp [x2d [string range $cmd 10 10]]
								if {[lsearch -regexp $set_commands "71 00 00 0[format %X $tmp] .. .. 00 00"] >= 0} {
									if {$resumeflag == -1} {
										set resumeflag $tmp
									}
									incr resumegprms($tmp) 2
									if {$vts == $cur_vts} {incr resumegprms($tmp) 1}
								}
							}
						}
						# If there is a GPRM in this PGC, search for conditional JUMPs and CALLs...
						# If one is found BEFORE the last line with a RSM, the GPRM used in the condition
						# may also be the resume flag
						set idxs [lsearch -all -regexp $::menus($vts,$lu,$pgc) {30 2[2-8] .. .. .. .. .. ..}]
						foreach idx $idxs {
							if {$idx < $last} {
								set found true
								set cmd [lindex $::menus($vts,$lu,$pgc) $idx]
								set tmp [x2d [string range $cmd 19 19]]
								if {[lsearch -regexp $set_commands "71 00 00 0[format %X $tmp] .. .. 00 00"] >= 0} {
									if {$resumeflag == -1} {
										set resumeflag $tmp
									}
									incr resumegprms($tmp) 2
									if {$vts == $cur_vts} {incr resumegprms($tmp) 1}
								}
							}
						}
					}
					# Also, search for RSM commands in buttons and BOVs.
					set numcells $::menus($vts,$lu,$pgc,numcells)
					if {$numcells == 0} continue
					set cellpostable $::menus($vts,$lu,$pgc,cellpostable)
					set butcmds {}
					for {set cell 0} {$cell < $numcells} {incr cell} {
						set idx [expr {$cell * 4}]
						set vobid [get_int cellpostable $idx 2]
						incr idx 3
						set cellid [get_int cellpostable $idx 1]
						if {[info exists ::buttons::curbutsinfo($vobid,$cellid,numbuts)]} {
							set numbuts $::buttons::curbutsinfo($vobid,$cellid,numbuts)
							if {$numbuts == 0} {continue}
							set numgroups $::buttons::curbutsinfo($vobid,$cellid,numgroups)
							if {$numgroups == 0} {continue}
							for {set butnum 1} {$butnum <= $numbuts} {incr butnum} {
								for {set group 1} {$group <= $numgroups} {incr group} {
									lappend butcmds [string toupper $::buttons::curbutsinfo($vobid,$cellid,$group,$butnum,command)]
								}
							}
						}
					}
					set idxs [lsearch -all -regexp $butcmds "20 .1 .. .. .. .. .. 10"]
					foreach idx $idxs {
						set found true
						# It is extremely rare to find a condition in a button command.
						# However, test it also...
						set cmd [lindex $butcmds $idx]
						if {[string range $cmd 3 3]!= 0} {
							set tmp [x2d [string range $cmd 10 10]]
							if {[lsearch -regexp $set_commands "71 00 00 0[format %X $tmp] .. .. 00 00"] >= 0} {
								set resumeflag $tmp
								incr resumegprms($tmp) 7
								if {$vts == $cur_vts} {incr resumegprms($tmp) 2}
							}
						}
					}
				}
			}
		}
	}
	show_msg ""
	if {$found} {
		set txt "Done.\n\nWARNING!  There are RSM commands in the menu domains."
		set txt "$txt\nIt is possible that a GPRM is used as a 'resume flag', to"
		set txt "$txt\nresume to the title from a menu.  If this GPRM is set by this"
		set txt "$txt\nmacro to the resume value in your target PGC, then when the RSM"
		set txt "$txt\ncommand will be executed, an infinite loop may begin."
		set txt "$txt\n\nTo avoid this problem, you should verify if a GPRM is associated"
		set txt "$txt\nwith the RSM commands, and if the same GPRM is set to a specific"
		set txt "$txt\nvalue somewhere before your main title begins, and resetted back"
		set txt "$txt\nto another value when the title has finished playing."
		set txt "$txt\nIn such case, you may need to delete the \"Set gprm(X) =(mov) Val\""
		set txt "$txt\nline added by PgcEdit in your target PGC."
		set txt "$txt\n\nThe RSM commands will now be displayed, and your target PGC"
		set txt "$txt\nselected."
		if {$resumeflag != -1} {
			set tmp 0
			set resumeflags {}
			for {set i 0} {$i < 16} {incr i} {
				if {$resumegprms($i) > 0} {
					lappend resumeflags $i
					if {$resumegprms($i) > $tmp} {
						set tmp $resumegprms($i)
						set resumeflag $i
					}
				}
			}
			if {[llength $resumeflags] > 1} {
				set txt "$txt\n\n\nPgcEdit has detected several GPRMs that may be used as the resume"
				set txt "$txt\nflag, as they are associated with IF tests around the RSM command."
				set txt "$txt\nThese registers are: gprm([string map {{ } {), gprm(}} $resumeflags])."
				set txt "$txt\n\nThe gprm($resumeflag) is probably the one used as the resume flag.  However,"
				set txt "$txt\nanother GPRM in the list may also be used.  If needed, try them all!"
			} elseif {[llength $resumeflags] == 1} {
				set txt "$txt\n\n\nPgcEdit has detected that GPRM $resumeflag is probably used as the resume"
				set txt "$txt\nflag, as it is associated with IF tests around the RSM command."
			}
			set txt "$txt\n\nIn case of endless loops, you should try to remove the first line"
			set txt "$txt\n\"Set gprm($resumeflag) =(mov) <any value>\"  in your target PGC."
			set txt "$txt\nThis line will be automatically selected by PgcEdit when you click OK..."
			set txt "$txt\n(Be sure to turn the Renumber GOTOs option ON before deleting the"
			set txt "$txt\nline, or change the Set command to NOP.)"
		}
		set txt "$txt\n\nTest the DVD carefully with your soft player before burning it!"
		set txt "$txt\nTry to go to the RSM commands, and see if there are endless loops."
		tk_messageBox -title $utiltitle -icon info -message $txt -parent .
		::main::display_pgc menu $cur_vts $cur_lu $cur_pgc
		if {$resumeflag != -1} {
			set cmds $::menus($cur_vts,$cur_lu,$cur_pgc)
			set idx [lsearch -regexp $cmds "71 00 00 0[format %X $resumeflag] .. .. 00 00"]
			if {$idx >= 0} {
				::main::set_listbox_selection $idx
			}
		}
		::utils::find_results "RSM" false false
	} else {
		set txt "Done.\n\nPlease trace again to your target PGC's playback to test!"
		show_msg $txt
		set txt "$txt\nBe sure to enable the preview and press the \"->>> Run\" button."
		set txt "$txt\nThe first thing that will be played should be your target PGC."
		set txt "$txt\nThen, save the DVD and test it carefully with your soft player."
		if {[info exists ::tr::breakpoints($cur_type,$cur_index,prepost)]} {
			set txt "$txt\n\nYou should not need the breakpoint on the target PGC anymore."
			set txt "$txt\nRemove it?  (recommended)"
			set r [tk_messageBox -title $utiltitle -icon question -message $txt -parent . \
					-type yesno -default yes]
			if {$r == "yes"} {
				unset -nocomplain ::tr::breakpoints($cur_type,$cur_index,prepost)
				::main::highlight_trace_breakpoints
			}
		} else {
			tk_messageBox -title $utiltitle -icon info -message $txt -parent .
		}
	}
}

}	;# end of namespace ::macros
